# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'main_window.ui'
##
## Created by: Qt User Interface Compiler version 6.7.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QAction, QBrush, QColor, QConicalGradient,
    QCursor, QFont, QFontDatabase, QGradient,
    QIcon, QImage, QKeySequence, QLinearGradient,
    QPainter, QPalette, QPixmap, QRadialGradient,
    QTransform)
from PySide6.QtWidgets import (QApplication, QGridLayout, QMainWindow, QMenu,
    QMenuBar, QSizePolicy, QSplitter, QStatusBar,
    QTabWidget, QToolBar, QVBoxLayout, QWidget)

from airunner.gui.widgets.canvas.canvas_widget import CanvasWidget
from airunner.gui.widgets.generator_form.generator_form_widget import GeneratorForm
from airunner.gui.widgets.nodegraph.node_graph_widget import NodeGraphWidget
from airunner.gui.widgets.tool_tab.tool_tab_widget import ToolTabWidget
import airunner.feather_rc

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        if not MainWindow.objectName():
            MainWindow.setObjectName(u"MainWindow")
        MainWindow.resize(1054, 778)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(MainWindow.sizePolicy().hasHeightForWidth())
        MainWindow.setSizePolicy(sizePolicy)
        MainWindow.setMinimumSize(QSize(0, 0))
        font = QFont()
        font.setPointSize(8)
        MainWindow.setFont(font)
        icon = QIcon()
        icon.addFile(u":/icons/icon_256.png", QSize(), QIcon.Normal, QIcon.Off)
        MainWindow.setWindowIcon(icon)
        MainWindow.setStyleSheet(u"")
        self.actionCopy = QAction(MainWindow)
        self.actionCopy.setObjectName(u"actionCopy")
        icon1 = QIcon()
        icon1.addFile(u":/light/icons/feather/light/copy.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actionCopy.setIcon(icon1)
        self.actionPaste = QAction(MainWindow)
        self.actionPaste.setObjectName(u"actionPaste")
        icon2 = QIcon()
        icon2.addFile(u":/light/icons/feather/light/clipboard.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actionPaste.setIcon(icon2)
        self.actionAbout = QAction(MainWindow)
        self.actionAbout.setObjectName(u"actionAbout")
        self.actionQuit = QAction(MainWindow)
        self.actionQuit.setObjectName(u"actionQuit")
        icon3 = QIcon()
        icon3.addFile(u":/light/icons/feather/light/x-circle.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actionQuit.setIcon(icon3)
        self.actionBug_report = QAction(MainWindow)
        self.actionBug_report.setObjectName(u"actionBug_report")
        icon4 = QIcon()
        icon4.addFile(u":/light/icons/feather/light/external-link.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actionBug_report.setIcon(icon4)
        self.actionReport_vulnerability = QAction(MainWindow)
        self.actionReport_vulnerability.setObjectName(u"actionReport_vulnerability")
        self.actionReport_vulnerability.setIcon(icon4)
        self.actionRotate_90_clockwise = QAction(MainWindow)
        self.actionRotate_90_clockwise.setObjectName(u"actionRotate_90_clockwise")
        icon5 = QIcon()
        icon5.addFile(u":/light/icons/feather/light/corner-right-down.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actionRotate_90_clockwise.setIcon(icon5)
        self.actionRotate_90_counter_clockwise = QAction(MainWindow)
        self.actionRotate_90_counter_clockwise.setObjectName(u"actionRotate_90_counter_clockwise")
        icon6 = QIcon()
        icon6.addFile(u":/light/icons/feather/light/corner-left-down.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actionRotate_90_counter_clockwise.setIcon(icon6)
        self.actionPrompt_Browser = QAction(MainWindow)
        self.actionPrompt_Browser.setObjectName(u"actionPrompt_Browser")
        icon7 = QIcon()
        icon7.addFile(u":/light/icons/feather/light/book-open.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actionPrompt_Browser.setIcon(icon7)
        self.actionClear_all_prompts = QAction(MainWindow)
        self.actionClear_all_prompts.setObjectName(u"actionClear_all_prompts")
        icon8 = QIcon()
        icon8.addFile(u":/light/icons/feather/light/delete.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actionClear_all_prompts.setIcon(icon8)
        self.actionCut = QAction(MainWindow)
        self.actionCut.setObjectName(u"actionCut")
        icon9 = QIcon()
        icon9.addFile(u":/light/icons/feather/light/scissors.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actionCut.setIcon(icon9)
        self.actionBrowse_AI_Runner_Path = QAction(MainWindow)
        self.actionBrowse_AI_Runner_Path.setObjectName(u"actionBrowse_AI_Runner_Path")
        icon10 = QIcon()
        icon10.addFile(u":/light/icons/feather/light/folder.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actionBrowse_AI_Runner_Path.setIcon(icon10)
        self.actionBrowse_Images_Path_2 = QAction(MainWindow)
        self.actionBrowse_Images_Path_2.setObjectName(u"actionBrowse_Images_Path_2")
        self.actionBrowse_Images_Path_2.setIcon(icon10)
        self.actionReset_Settings_2 = QAction(MainWindow)
        self.actionReset_Settings_2.setObjectName(u"actionReset_Settings_2")
        icon11 = QIcon()
        icon11.addFile(u":/light/icons/feather/light/refresh-cw.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actionReset_Settings_2.setIcon(icon11)
        self.actionSafety_Checker = QAction(MainWindow)
        self.actionSafety_Checker.setObjectName(u"actionSafety_Checker")
        self.actionSafety_Checker.setCheckable(True)
        self.actionSafety_Checker.setChecked(False)
        icon12 = QIcon()
        icon12.addFile(u":/light/icons/feather/light/slash.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actionSafety_Checker.setIcon(icon12)
        self.actionRun_setup_wizard_2 = QAction(MainWindow)
        self.actionRun_setup_wizard_2.setObjectName(u"actionRun_setup_wizard_2")
        icon13 = QIcon()
        icon13.addFile(u":/light/icons/feather/light/zap.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actionRun_setup_wizard_2.setIcon(icon13)
        self.actionStats = QAction(MainWindow)
        self.actionStats.setObjectName(u"actionStats")
        icon14 = QIcon()
        icon14.addFile(u":/light/icons/feather/light/activity.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actionStats.setIcon(icon14)
        self.actionSettings = QAction(MainWindow)
        self.actionSettings.setObjectName(u"actionSettings")
        icon15 = QIcon()
        icon15.addFile(u":/light/icons/feather/light/settings.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actionSettings.setIcon(icon15)
        self.actionToggle_LLM = QAction(MainWindow)
        self.actionToggle_LLM.setObjectName(u"actionToggle_LLM")
        self.actionToggle_LLM.setCheckable(True)
        self.actionToggle_LLM.setChecked(False)
        icon16 = QIcon()
        icon16.addFile(u":/light/icons/feather/light/cpu.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actionToggle_LLM.setIcon(icon16)
        self.actionToggle_Text_to_Speech = QAction(MainWindow)
        self.actionToggle_Text_to_Speech.setObjectName(u"actionToggle_Text_to_Speech")
        self.actionToggle_Text_to_Speech.setCheckable(True)
        icon17 = QIcon()
        icon17.addFile(u":/light/icons/feather/light/speaker.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actionToggle_Text_to_Speech.setIcon(icon17)
        self.actionToggle_Speech_to_Text = QAction(MainWindow)
        self.actionToggle_Speech_to_Text.setObjectName(u"actionToggle_Speech_to_Text")
        self.actionToggle_Speech_to_Text.setCheckable(True)
        icon18 = QIcon()
        icon18.addFile(u":/light/icons/feather/light/mic.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actionToggle_Speech_to_Text.setIcon(icon18)
        self.actionToggle_Stable_Diffusion = QAction(MainWindow)
        self.actionToggle_Stable_Diffusion.setObjectName(u"actionToggle_Stable_Diffusion")
        self.actionToggle_Stable_Diffusion.setCheckable(True)
        icon19 = QIcon()
        icon19.addFile(u":/light/icons/feather/light/image.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actionToggle_Stable_Diffusion.setIcon(icon19)
        self.actionToggle_Controlnet = QAction(MainWindow)
        self.actionToggle_Controlnet.setObjectName(u"actionToggle_Controlnet")
        self.actionToggle_Controlnet.setCheckable(True)
        self.actionToggle_Controlnet.setChecked(False)
        icon20 = QIcon()
        icon20.addFile(u":/light/icons/feather/light/crosshair.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actionToggle_Controlnet.setIcon(icon20)
        self.actionToggle_Active_Grid_Area = QAction(MainWindow)
        self.actionToggle_Active_Grid_Area.setObjectName(u"actionToggle_Active_Grid_Area")
        self.actionToggle_Active_Grid_Area.setCheckable(True)
        icon21 = QIcon()
        icon21.addFile(u":/light/icons/feather/light/object-selected-icon.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actionToggle_Active_Grid_Area.setIcon(icon21)
        self.actionToggle_Brush = QAction(MainWindow)
        self.actionToggle_Brush.setObjectName(u"actionToggle_Brush")
        self.actionToggle_Brush.setCheckable(True)
        icon22 = QIcon()
        icon22.addFile(u":/light/icons/feather/light/pencil-icon.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actionToggle_Brush.setIcon(icon22)
        self.actionToggle_Grid = QAction(MainWindow)
        self.actionToggle_Grid.setObjectName(u"actionToggle_Grid")
        self.actionToggle_Grid.setCheckable(True)
        icon23 = QIcon()
        icon23.addFile(u":/light/icons/feather/light/grid.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actionToggle_Grid.setIcon(icon23)
        self.actionToggle_Eraser = QAction(MainWindow)
        self.actionToggle_Eraser.setObjectName(u"actionToggle_Eraser")
        self.actionToggle_Eraser.setCheckable(True)
        icon24 = QIcon()
        icon24.addFile(u":/light/icons/feather/light/eraser-icon.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actionToggle_Eraser.setIcon(icon24)
        self.artActionNew = QAction(MainWindow)
        self.artActionNew.setObjectName(u"artActionNew")
        icon25 = QIcon()
        icon25.addFile(u":/light/icons/feather/light/plus-circle.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.artActionNew.setIcon(icon25)
        self.actionImport_image = QAction(MainWindow)
        self.actionImport_image.setObjectName(u"actionImport_image")
        icon26 = QIcon()
        icon26.addFile(u":/light/icons/feather/light/download.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actionImport_image.setIcon(icon26)
        self.actionExport_image_button = QAction(MainWindow)
        self.actionExport_image_button.setObjectName(u"actionExport_image_button")
        icon27 = QIcon()
        icon27.addFile(u":/light/icons/feather/light/upload.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actionExport_image_button.setIcon(icon27)
        self.actionNew_Conversation = QAction(MainWindow)
        self.actionNew_Conversation.setObjectName(u"actionNew_Conversation")
        icon28 = QIcon()
        icon28.addFile(u":/light/icons/feather/light/message-circle.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actionNew_Conversation.setIcon(icon28)
        self.actionDelete_conversation = QAction(MainWindow)
        self.actionDelete_conversation.setObjectName(u"actionDelete_conversation")
        icon29 = QIcon()
        icon29.addFile(u":/light/icons/feather/light/trash-2.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actionDelete_conversation.setIcon(icon29)
        self.actionDiscord = QAction(MainWindow)
        self.actionDiscord.setObjectName(u"actionDiscord")
        icon30 = QIcon()
        icon30.addFile(u":/light/icons/feather/light/message-square.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actionDiscord.setIcon(icon30)
        self.actionCollapse_to_system_tray = QAction(MainWindow)
        self.actionCollapse_to_system_tray.setObjectName(u"actionCollapse_to_system_tray")
        self.actionCollapse_to_system_tray.setCheckable(True)
        self.actionCollapse_to_system_tray.setIcon(icon3)
        self.workflow_actionRun = QAction(MainWindow)
        self.workflow_actionRun.setObjectName(u"workflow_actionRun")
        icon31 = QIcon()
        icon31.addFile(u":/light/icons/feather/light/play.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.workflow_actionRun.setIcon(icon31)
        self.workflow_actionPause = QAction(MainWindow)
        self.workflow_actionPause.setObjectName(u"workflow_actionPause")
        icon32 = QIcon()
        icon32.addFile(u":/light/icons/feather/light/pause-circle.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.workflow_actionPause.setIcon(icon32)
        self.workflow_actionStop = QAction(MainWindow)
        self.workflow_actionStop.setObjectName(u"workflow_actionStop")
        icon33 = QIcon()
        icon33.addFile(u":/light/icons/feather/light/stop-circle.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.workflow_actionStop.setIcon(icon33)
        self.workflow_actionSave = QAction(MainWindow)
        self.workflow_actionSave.setObjectName(u"workflow_actionSave")
        icon34 = QIcon()
        icon34.addFile(u":/light/icons/feather/light/save.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.workflow_actionSave.setIcon(icon34)
        self.workflow_actionClear = QAction(MainWindow)
        self.workflow_actionClear.setObjectName(u"workflow_actionClear")
        self.workflow_actionClear.setIcon(icon29)
        self.workflow_actionEdit = QAction(MainWindow)
        self.workflow_actionEdit.setObjectName(u"workflow_actionEdit")
        icon35 = QIcon()
        icon35.addFile(u":/light/icons/feather/light/edit.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.workflow_actionEdit.setIcon(icon35)
        self.workflow_actionOpen = QAction(MainWindow)
        self.workflow_actionOpen.setObjectName(u"workflow_actionOpen")
        self.workflow_actionOpen.setIcon(icon10)
        self.actionSave_As = QAction(MainWindow)
        self.actionSave_As.setObjectName(u"actionSave_As")
        self.actionSave_As.setIcon(icon34)
        self.centralwidget = QWidget(MainWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        self.centralwidget.setEnabled(True)
        sizePolicy.setHeightForWidth(self.centralwidget.sizePolicy().hasHeightForWidth())
        self.centralwidget.setSizePolicy(sizePolicy)
        self.centralwidget.setMinimumSize(QSize(0, 0))
        self.centralwidget.setLayoutDirection(Qt.LayoutDirection.LeftToRight)
        self.gridLayout_3 = QGridLayout(self.centralwidget)
        self.gridLayout_3.setSpacing(0)
        self.gridLayout_3.setObjectName(u"gridLayout_3")
        self.gridLayout_3.setContentsMargins(0, 0, 0, 0)
        self.mode_tab_widget = QWidget(self.centralwidget)
        self.mode_tab_widget.setObjectName(u"mode_tab_widget")
        self.mode_tab_widget.setAutoFillBackground(False)
        self.mode_tab_widget.setStyleSheet(u"QTabWidget#mode_tab_widget::pane { border: 0; background: transparent; }")
        self.gridLayout = QGridLayout(self.mode_tab_widget)
        self.gridLayout.setSpacing(0)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setContentsMargins(0, 1, 0, 0)
        self.main_window_splitter = QSplitter(self.mode_tab_widget)
        self.main_window_splitter.setObjectName(u"main_window_splitter")
        self.main_window_splitter.setOrientation(Qt.Orientation.Horizontal)
        self.generator_widget = GeneratorForm(self.main_window_splitter)
        self.generator_widget.setObjectName(u"generator_widget")
        self.generator_widget.setEnabled(True)
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Preferred)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.generator_widget.sizePolicy().hasHeightForWidth())
        self.generator_widget.setSizePolicy(sizePolicy1)
        self.generator_widget.setMinimumSize(QSize(200, 0))
        self.main_window_splitter.addWidget(self.generator_widget)
        self.center_widget = QWidget(self.main_window_splitter)
        self.center_widget.setObjectName(u"center_widget")
        self.center_widget.setMinimumSize(QSize(0, 0))
        self.gridLayout_4 = QGridLayout(self.center_widget)
        self.gridLayout_4.setObjectName(u"gridLayout_4")
        self.gridLayout_4.setContentsMargins(0, 0, 0, 0)
        self.center_tab_container = QTabWidget(self.center_widget)
        self.center_tab_container.setObjectName(u"center_tab_container")
        self.center_tab_container.setTabsClosable(False)
        self.center_tab_container.setMovable(False)
        self.art_tab = QWidget()
        self.art_tab.setObjectName(u"art_tab")
        self.gridLayout_2 = QGridLayout(self.art_tab)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.gridLayout_2.setContentsMargins(0, 0, 0, 0)
        self.canvas = CanvasWidget(self.art_tab)
        self.canvas.setObjectName(u"canvas")
        sizePolicy2 = QSizePolicy(QSizePolicy.Policy.Ignored, QSizePolicy.Policy.MinimumExpanding)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.canvas.sizePolicy().hasHeightForWidth())
        self.canvas.setSizePolicy(sizePolicy2)
        self.canvas.setMinimumSize(QSize(0, 0))
        self.canvas.setMaximumSize(QSize(16777215, 16777215))
        self.canvas.setCursor(QCursor(Qt.ArrowCursor))
        self.canvas.setAcceptDrops(True)

        self.gridLayout_2.addWidget(self.canvas, 0, 0, 1, 1)

        self.center_tab_container.addTab(self.art_tab, "")
        self.agent_workflow_tab = QWidget()
        self.agent_workflow_tab.setObjectName(u"agent_workflow_tab")
        self.gridLayout_5 = QGridLayout(self.agent_workflow_tab)
        self.gridLayout_5.setObjectName(u"gridLayout_5")
        self.gridLayout_5.setContentsMargins(0, 0, 0, 0)
        self.graph = NodeGraphWidget(self.agent_workflow_tab)
        self.graph.setObjectName(u"graph")

        self.gridLayout_5.addWidget(self.graph, 0, 0, 1, 1)

        self.center_tab_container.addTab(self.agent_workflow_tab, "")

        self.gridLayout_4.addWidget(self.center_tab_container, 0, 0, 1, 1)

        self.main_window_splitter.addWidget(self.center_widget)
        self.panel_container = QWidget(self.main_window_splitter)
        self.panel_container.setObjectName(u"panel_container")
        self.verticalLayout_2 = QVBoxLayout(self.panel_container)
        self.verticalLayout_2.setSpacing(0)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.tool_tab_widget = ToolTabWidget(self.panel_container)
        self.tool_tab_widget.setObjectName(u"tool_tab_widget")
        self.tool_tab_widget.setMinimumSize(QSize(50, 0))

        self.verticalLayout_2.addWidget(self.tool_tab_widget)

        self.main_window_splitter.addWidget(self.panel_container)

        self.gridLayout.addWidget(self.main_window_splitter, 0, 0, 1, 1)


        self.gridLayout_3.addWidget(self.mode_tab_widget, 0, 0, 1, 1)

        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QMenuBar(MainWindow)
        self.menubar.setObjectName(u"menubar")
        self.menubar.setGeometry(QRect(0, 0, 1054, 23))
        font1 = QFont()
        font1.setPointSize(11)
        self.menubar.setFont(font1)
        self.menuFile = QMenu(self.menubar)
        self.menuFile.setObjectName(u"menuFile")
        self.menuFile.setTearOffEnabled(False)
        self.menuFile.setSeparatorsCollapsible(False)
        self.menuArt = QMenu(self.menuFile)
        self.menuArt.setObjectName(u"menuArt")
        self.menuArt.setIcon(icon19)
        self.menuChat = QMenu(self.menuFile)
        self.menuChat.setObjectName(u"menuChat")
        self.menuChat.setIcon(icon28)
        self.menuWorkflow = QMenu(self.menuFile)
        self.menuWorkflow.setObjectName(u"menuWorkflow")
        icon36 = QIcon()
        icon36.addFile(u":/light/icons/feather/light/codesandbox.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.menuWorkflow.setIcon(icon36)
        self.menuEdit = QMenu(self.menubar)
        self.menuEdit.setObjectName(u"menuEdit")
        self.menuFilters = QMenu(self.menubar)
        self.menuFilters.setObjectName(u"menuFilters")
        self.menuAbout = QMenu(self.menubar)
        self.menuAbout.setObjectName(u"menuAbout")
        self.menuAbout.setTearOffEnabled(False)
        self.menuImage = QMenu(self.menubar)
        self.menuImage.setObjectName(u"menuImage")
        self.menuView = QMenu(self.menubar)
        self.menuView.setObjectName(u"menuView")
        self.menuTools = QMenu(self.menubar)
        self.menuTools.setObjectName(u"menuTools")
        self.menuStable_Diffusion = QMenu(self.menuTools)
        self.menuStable_Diffusion.setObjectName(u"menuStable_Diffusion")
        self.menuStable_Diffusion.setIcon(icon19)
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QStatusBar(MainWindow)
        self.statusbar.setObjectName(u"statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.model_toolbar = QToolBar(MainWindow)
        self.model_toolbar.setObjectName(u"model_toolbar")
        self.model_toolbar.setLayoutDirection(Qt.LayoutDirection.RightToLeft)
        self.model_toolbar.setMovable(True)
        self.model_toolbar.setOrientation(Qt.Orientation.Horizontal)
        self.model_toolbar.setIconSize(QSize(18, 18))
        self.model_toolbar.setFloatable(False)
        MainWindow.addToolBar(Qt.ToolBarArea.BottomToolBarArea, self.model_toolbar)

        self.menubar.addAction(self.menuFile.menuAction())
        self.menubar.addAction(self.menuEdit.menuAction())
        self.menubar.addAction(self.menuView.menuAction())
        self.menubar.addAction(self.menuImage.menuAction())
        self.menubar.addAction(self.menuFilters.menuAction())
        self.menubar.addAction(self.menuTools.menuAction())
        self.menubar.addAction(self.menuAbout.menuAction())
        self.menuFile.addAction(self.menuArt.menuAction())
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.menuChat.menuAction())
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.menuWorkflow.menuAction())
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionReset_Settings_2)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionQuit)
        self.menuArt.addAction(self.artActionNew)
        self.menuArt.addAction(self.actionImport_image)
        self.menuArt.addAction(self.actionExport_image_button)
        self.menuChat.addAction(self.actionNew_Conversation)
        self.menuChat.addAction(self.actionDelete_conversation)
        self.menuWorkflow.addAction(self.workflow_actionRun)
        self.menuWorkflow.addAction(self.workflow_actionPause)
        self.menuWorkflow.addAction(self.workflow_actionStop)
        self.menuWorkflow.addSeparator()
        self.menuWorkflow.addAction(self.workflow_actionSave)
        self.menuWorkflow.addAction(self.actionSave_As)
        self.menuWorkflow.addAction(self.workflow_actionClear)
        self.menuWorkflow.addAction(self.workflow_actionEdit)
        self.menuWorkflow.addAction(self.workflow_actionOpen)
        self.menuEdit.addSeparator()
        self.menuEdit.addAction(self.actionCut)
        self.menuEdit.addAction(self.actionCopy)
        self.menuEdit.addAction(self.actionPaste)
        self.menuEdit.addSeparator()
        self.menuEdit.addAction(self.actionRotate_90_clockwise)
        self.menuEdit.addAction(self.actionRotate_90_counter_clockwise)
        self.menuEdit.addSeparator()
        self.menuEdit.addAction(self.actionClear_all_prompts)
        self.menuEdit.addSeparator()
        self.menuEdit.addAction(self.actionSettings)
        self.menuFilters.addSeparator()
        self.menuAbout.addAction(self.actionAbout)
        self.menuAbout.addAction(self.actionBug_report)
        self.menuAbout.addAction(self.actionReport_vulnerability)
        self.menuAbout.addSeparator()
        self.menuAbout.addAction(self.actionDiscord)
        self.menuImage.addSeparator()
        self.menuImage.addSeparator()
        self.menuImage.addAction(self.actionToggle_Active_Grid_Area)
        self.menuImage.addAction(self.actionToggle_Brush)
        self.menuImage.addAction(self.actionToggle_Eraser)
        self.menuImage.addAction(self.actionToggle_Grid)
        self.menuView.addAction(self.actionPrompt_Browser)
        self.menuView.addSeparator()
        self.menuView.addAction(self.actionBrowse_AI_Runner_Path)
        self.menuView.addSeparator()
        self.menuView.addAction(self.actionBrowse_Images_Path_2)
        self.menuView.addSeparator()
        self.menuTools.addAction(self.menuStable_Diffusion.menuAction())
        self.menuTools.addSeparator()
        self.menuTools.addAction(self.actionToggle_LLM)
        self.menuTools.addAction(self.actionToggle_Text_to_Speech)
        self.menuTools.addAction(self.actionToggle_Speech_to_Text)
        self.menuTools.addSeparator()
        self.menuTools.addAction(self.actionStats)
        self.menuTools.addSeparator()
        self.menuTools.addAction(self.actionRun_setup_wizard_2)
        self.menuTools.addSeparator()
        self.menuTools.addAction(self.actionCollapse_to_system_tray)
        self.menuStable_Diffusion.addAction(self.actionSafety_Checker)
        self.menuStable_Diffusion.addAction(self.actionToggle_Stable_Diffusion)
        self.menuStable_Diffusion.addAction(self.actionToggle_Controlnet)
        self.model_toolbar.addAction(self.actionSettings)
        self.model_toolbar.addSeparator()
        self.model_toolbar.addAction(self.actionToggle_Speech_to_Text)
        self.model_toolbar.addAction(self.actionToggle_Text_to_Speech)
        self.model_toolbar.addAction(self.actionToggle_LLM)
        self.model_toolbar.addSeparator()
        self.model_toolbar.addAction(self.actionToggle_Controlnet)
        self.model_toolbar.addAction(self.actionToggle_Stable_Diffusion)

        self.retranslateUi(MainWindow)
        self.actionAbout.triggered.connect(MainWindow.action_show_about_window)
        self.actionQuit.triggered.connect(MainWindow.action_quit_triggered)
        self.actionPaste.triggered.connect(MainWindow.action_paste_image_triggered)
        self.actionCopy.triggered.connect(MainWindow.action_copy_image_triggered)
        self.actionCut.triggered.connect(MainWindow.action_cut_image_triggered)
        self.actionRotate_90_clockwise.triggered.connect(MainWindow.action_rotate_90_clockwise_triggered)
        self.actionRotate_90_counter_clockwise.triggered.connect(MainWindow.action_rotate_90_counterclockwise_triggered)
        self.actionPrompt_Browser.triggered.connect(MainWindow.action_show_prompt_browser_triggered)
        self.actionClear_all_prompts.triggered.connect(MainWindow.action_clear_all_prompts_triggered)
        self.actionBrowse_AI_Runner_Path.triggered.connect(MainWindow.action_triggered_browse_ai_runner_path)
        self.actionBrowse_Images_Path_2.triggered.connect(MainWindow.action_show_images_path)
        self.actionBug_report.triggered.connect(MainWindow.action_open_bug_report)
        self.actionReport_vulnerability.triggered.connect(MainWindow.action_open_vulnerability_report)
        self.actionReset_Settings_2.triggered.connect(MainWindow.action_reset_settings)
        self.actionSafety_Checker.toggled.connect(MainWindow.action_toggle_nsfw_filter_triggered)
        self.actionRun_setup_wizard_2.triggered.connect(MainWindow.action_run_setup_wizard_clicked)
        self.actionStats.triggered.connect(MainWindow.action_stats_triggered)
        self.actionSettings.triggered.connect(MainWindow.action_show_settings)
        self.actionToggle_Controlnet.toggled.connect(MainWindow.action_controlnet_toggled)
        self.actionToggle_Stable_Diffusion.toggled.connect(MainWindow.action_image_generator_toggled)
        self.actionToggle_Text_to_Speech.toggled.connect(MainWindow.tts_button_toggled)
        self.actionToggle_Speech_to_Text.toggled.connect(MainWindow.stt_button_toggled)
        self.actionToggle_Grid.toggled.connect(MainWindow.action_toggle_grid)
        self.actionToggle_Active_Grid_Area.toggled.connect(MainWindow.action_toggle_active_grid_area)
        self.actionToggle_Brush.toggled.connect(MainWindow.action_toggle_brush)
        self.actionToggle_Eraser.toggled.connect(MainWindow.action_toggle_eraser)
        self.actionToggle_LLM.toggled.connect(MainWindow.action_toggle_llm)
        self.artActionNew.triggered.connect(MainWindow.action_new_document_triggered)
        self.actionNew_Conversation.triggered.connect(MainWindow.action_conversation_triggered_new)
        self.actionDelete_conversation.triggered.connect(MainWindow.action_conversation_triggered_delete)
        self.actionImport_image.triggered.connect(MainWindow.action_import_image_triggered)
        self.actionExport_image_button.triggered.connect(MainWindow.action_export_image_triggered)
        self.actionDiscord.triggered.connect(MainWindow.action_open_discord)
        self.actionCollapse_to_system_tray.toggled.connect(MainWindow.action_toggle_close_to_system_tray)
        self.workflow_actionRun.triggered.connect(MainWindow.workflow_actionRun_triggered)
        self.menubar.triggered.connect(MainWindow.workflow_actionRun_triggered)
        self.workflow_actionClear.triggered.connect(MainWindow.workflow_actionClear_triggered)
        self.workflow_actionEdit.triggered.connect(MainWindow.workflow_actionEdit_triggered)
        self.workflow_actionOpen.triggered.connect(MainWindow.workflow_actionOpen_triggered)
        self.workflow_actionPause.triggered.connect(MainWindow.workflow_actionPause_triggered)
        self.workflow_actionSave.triggered.connect(MainWindow.workflow_actionSave_triggered)
        self.workflow_actionStop.triggered.connect(MainWindow.workflow_actionStop_triggered)

        self.center_tab_container.setCurrentIndex(0)


        QMetaObject.connectSlotsByName(MainWindow)
    # setupUi

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(QCoreApplication.translate("MainWindow", u"AI Runner", None))
        self.actionCopy.setText(QCoreApplication.translate("MainWindow", u"Copy", None))
#if QT_CONFIG(shortcut)
        self.actionCopy.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+C", None))
#endif // QT_CONFIG(shortcut)
        self.actionPaste.setText(QCoreApplication.translate("MainWindow", u"Paste", None))
#if QT_CONFIG(shortcut)
        self.actionPaste.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+V", None))
#endif // QT_CONFIG(shortcut)
        self.actionAbout.setText(QCoreApplication.translate("MainWindow", u"About", None))
        self.actionQuit.setText(QCoreApplication.translate("MainWindow", u"Quit", None))
#if QT_CONFIG(shortcut)
        self.actionQuit.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+Q", None))
#endif // QT_CONFIG(shortcut)
        self.actionBug_report.setText(QCoreApplication.translate("MainWindow", u"Bug report", None))
        self.actionReport_vulnerability.setText(QCoreApplication.translate("MainWindow", u"Report vulnerability", None))
        self.actionRotate_90_clockwise.setText(QCoreApplication.translate("MainWindow", u"Rotate 90\u00b0 clockwise", None))
#if QT_CONFIG(shortcut)
        self.actionRotate_90_clockwise.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+R", None))
#endif // QT_CONFIG(shortcut)
        self.actionRotate_90_counter_clockwise.setText(QCoreApplication.translate("MainWindow", u"Rotate 90\u00b0 counter clockwise", None))
#if QT_CONFIG(shortcut)
        self.actionRotate_90_counter_clockwise.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+Shift+R", None))
#endif // QT_CONFIG(shortcut)
        self.actionPrompt_Browser.setText(QCoreApplication.translate("MainWindow", u"Prompt browser", None))
        self.actionClear_all_prompts.setText(QCoreApplication.translate("MainWindow", u"Clear all prompts", None))
#if QT_CONFIG(tooltip)
        self.actionClear_all_prompts.setToolTip(QCoreApplication.translate("MainWindow", u"Remove text from all prompts and negative prompts", None))
#endif // QT_CONFIG(tooltip)
        self.actionCut.setText(QCoreApplication.translate("MainWindow", u"Cut", None))
#if QT_CONFIG(shortcut)
        self.actionCut.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+X", None))
#endif // QT_CONFIG(shortcut)
        self.actionBrowse_AI_Runner_Path.setText(QCoreApplication.translate("MainWindow", u"Browse AI Runner Path", None))
        self.actionBrowse_Images_Path_2.setText(QCoreApplication.translate("MainWindow", u"Browse Images", None))
        self.actionReset_Settings_2.setText(QCoreApplication.translate("MainWindow", u"Reset Settings", None))
        self.actionSafety_Checker.setText(QCoreApplication.translate("MainWindow", u"Safety Checker", None))
        self.actionRun_setup_wizard_2.setText(QCoreApplication.translate("MainWindow", u"Run Setup Wizard", None))
        self.actionStats.setText(QCoreApplication.translate("MainWindow", u"Stats", None))
        self.actionSettings.setText(QCoreApplication.translate("MainWindow", u"Preferences", None))
        self.actionToggle_LLM.setText(QCoreApplication.translate("MainWindow", u"Toggle LLM", None))
        self.actionToggle_Text_to_Speech.setText(QCoreApplication.translate("MainWindow", u"Toggle Text to Speech", None))
        self.actionToggle_Speech_to_Text.setText(QCoreApplication.translate("MainWindow", u"Toggle Speech to Text", None))
        self.actionToggle_Stable_Diffusion.setText(QCoreApplication.translate("MainWindow", u"Toggle Stable Diffusion", None))
        self.actionToggle_Controlnet.setText(QCoreApplication.translate("MainWindow", u"Toggle Controlnet", None))
        self.actionToggle_Active_Grid_Area.setText(QCoreApplication.translate("MainWindow", u"Toggle Active Grid Area", None))
        self.actionToggle_Brush.setText(QCoreApplication.translate("MainWindow", u"Toggle Brush", None))
        self.actionToggle_Grid.setText(QCoreApplication.translate("MainWindow", u"Toggle Grid", None))
        self.actionToggle_Eraser.setText(QCoreApplication.translate("MainWindow", u"Toggle Eraser", None))
        self.artActionNew.setText(QCoreApplication.translate("MainWindow", u"New", None))
#if QT_CONFIG(shortcut)
        self.artActionNew.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+N", None))
#endif // QT_CONFIG(shortcut)
        self.actionImport_image.setText(QCoreApplication.translate("MainWindow", u"Import image", None))
#if QT_CONFIG(shortcut)
        self.actionImport_image.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+I", None))
#endif // QT_CONFIG(shortcut)
        self.actionExport_image_button.setText(QCoreApplication.translate("MainWindow", u"Export image", None))
#if QT_CONFIG(shortcut)
        self.actionExport_image_button.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+E", None))
#endif // QT_CONFIG(shortcut)
        self.actionNew_Conversation.setText(QCoreApplication.translate("MainWindow", u"New Conversation", None))
        self.actionDelete_conversation.setText(QCoreApplication.translate("MainWindow", u"Delete conversation", None))
        self.actionDiscord.setText(QCoreApplication.translate("MainWindow", u"Discord", None))
        self.actionCollapse_to_system_tray.setText(QCoreApplication.translate("MainWindow", u"Close to System Tray", None))
        self.workflow_actionRun.setText(QCoreApplication.translate("MainWindow", u"Run", None))
        self.workflow_actionPause.setText(QCoreApplication.translate("MainWindow", u"Pause", None))
        self.workflow_actionStop.setText(QCoreApplication.translate("MainWindow", u"Stop", None))
        self.workflow_actionSave.setText(QCoreApplication.translate("MainWindow", u"Save", None))
        self.workflow_actionClear.setText(QCoreApplication.translate("MainWindow", u"Clear", None))
        self.workflow_actionEdit.setText(QCoreApplication.translate("MainWindow", u"Edit", None))
        self.workflow_actionOpen.setText(QCoreApplication.translate("MainWindow", u"Load", None))
        self.actionSave_As.setText(QCoreApplication.translate("MainWindow", u"Save As", None))
        self.center_tab_container.setTabText(self.center_tab_container.indexOf(self.art_tab), QCoreApplication.translate("MainWindow", u"Art", None))
        self.center_tab_container.setTabText(self.center_tab_container.indexOf(self.agent_workflow_tab), QCoreApplication.translate("MainWindow", u"Agent Workflow", None))
        self.menuFile.setTitle(QCoreApplication.translate("MainWindow", u"File", None))
        self.menuArt.setTitle(QCoreApplication.translate("MainWindow", u"Art", None))
        self.menuChat.setTitle(QCoreApplication.translate("MainWindow", u"Chat", None))
        self.menuWorkflow.setTitle(QCoreApplication.translate("MainWindow", u"Workflow", None))
        self.menuEdit.setTitle(QCoreApplication.translate("MainWindow", u"Edit", None))
        self.menuFilters.setTitle(QCoreApplication.translate("MainWindow", u"Filters", None))
        self.menuAbout.setTitle(QCoreApplication.translate("MainWindow", u"Help", None))
        self.menuImage.setTitle(QCoreApplication.translate("MainWindow", u"Canvas", None))
        self.menuView.setTitle(QCoreApplication.translate("MainWindow", u"View", None))
        self.menuTools.setTitle(QCoreApplication.translate("MainWindow", u"Tools", None))
        self.menuStable_Diffusion.setTitle(QCoreApplication.translate("MainWindow", u"Stable Diffusion", None))
        self.model_toolbar.setWindowTitle(QCoreApplication.translate("MainWindow", u"toolBar_2", None))
    # retranslateUi




# Node representing a nested Workflow
from airunner.gui.widgets.nodegraph.nodes.core.base_core_node import (
    BaseCoreNode,
)


class WorkflowNode(BaseCoreNode):
    NODE_NAME = "Workflow"  # Default name

    def __init__(self):
        super().__init__()
        # Ports for triggering and receiving results from the nested workflow
        self.add_input("start_flow")
        self.add_output("flow_complete")
        # Add a property to store the ID or name of the nested workflow
        self.add_text_input(
            "nested_workflow_id", "Workflow ID/Name", tab="widgets"
        )

    # Override execute for Workflow specific logic (placeholder)
    def execute(self, input_data):
        nested_workflow_id = self.get_property("nested_workflow_id")
        start_data = input_data.get("start_flow", None)
        print(
            f"Executing nested Workflow: {nested_workflow_id} with start data: {start_data}"
        )
        # Dummy logic: Load and execute the nested workflow based on nested_workflow_id.
        # For now, just pass data through.
        output_data = f"Workflow '{nested_workflow_id}' completed, result from: {start_data}"
        return {
            "flow_complete": output_data
        }  # Return data for the 'flow_complete' port



from typing import Optional
from PySide6.QtCore import QPoint
from PySide6.QtGui import QPainter
from PySide6.QtWidgets import QGraphicsItem, QGraphicsPixmapItem

from airunner.enums import CanvasToolName
from airunner.utils.application.mediator_mixin import MediatorMixin
from airunner.utils.application import snap_to_grid
from airunner.gui.windows.main.settings_mixin import SettingsMixin
from airunner.utils.settings import get_qsettings


class DraggablePixmap(
    MediatorMixin,
    SettingsMixin,
    QGraphicsPixmapItem,
):
    def __init__(self, pixmap):
        super().__init__(pixmap)
        self.pixmap = pixmap
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsMovable, True)
        self.last_pos = QPoint(0, 0)
        self.save = False
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsMovable, True)
        pos = self.drawing_pad_settings.pos
        self.settings = get_qsettings()
        # Threshold for snapping (percentage of cell size)
        self.snap_threshold = 0.5
        self.setPos(QPoint(pos[0], pos[1]))

    @property
    def canvas_offset_x(self) -> float:
        return self.settings.value("canvas_offset_x", 0, type=float)

    @property
    def canvas_offset_y(self) -> float:
        return self.settings.value("canvas_offset_y", 0, type=float)

    @property
    def current_tool(self):
        return CanvasToolName(self.application_settings.current_tool)

    def mouseMoveEvent(self, event):
        if self.current_tool not in [
            CanvasToolName.ACTIVE_GRID_AREA,
        ]:
            return
        super().mouseMoveEvent(event)
        self.snap_to_grid_when_close()

    def mouseReleaseEvent(self, event):
        if self.current_tool in [
            CanvasToolName.ACTIVE_GRID_AREA,
        ]:
            # Pass the current position for saving instead of None values
            current_x = int(self.x())
            current_y = int(self.y())
            self.update_position(x=current_x, y=current_y, save=True)
        super().mouseReleaseEvent(event)

    def snap_to_grid_when_close(self, x=None, y=None):
        """Only snap to grid when close enough to grid lines"""
        if x is None:
            x = int(self.x())
        if y is None:
            y = int(self.y())

        # Get the absolute coordinates (including canvas offset)
        abs_x = x + self.canvas_offset_x
        abs_y = y + self.canvas_offset_y

        if self.grid_settings.snap_to_grid:
            cell_size = self.grid_settings.cell_size

            if cell_size <= 0:  # Safety check
                return

            # Calculate threshold distance in pixels
            threshold_dist = cell_size * self.snap_threshold

            # Calculate distance to nearest grid line in x and y directions
            x_mod = abs_x % cell_size
            y_mod = abs_y % cell_size

            # Get distance to closest grid line
            x_dist = min(x_mod, cell_size - x_mod)
            y_dist = min(y_mod, cell_size - y_mod)

            # Determine if we should snap in each direction
            snap_x = x_dist <= threshold_dist
            snap_y = y_dist <= threshold_dist

            # Only snap the axes that are close to grid lines
            if snap_x or snap_y:
                # Get the potential snapped position
                snapped_x, snapped_y = snap_to_grid(
                    self.grid_settings,
                    abs_x,
                    abs_y,
                    False,  # Use rounding for better visual centering
                )

                # Only apply snapping to the axes that should be snapped
                if snap_x:
                    x = snapped_x - self.canvas_offset_x
                if snap_y:
                    y = snapped_y - self.canvas_offset_y

        self.update_position(x, y, False)  # Don't save during dragging

    def update_position(
        self,
        x: Optional[int] = None,
        y: Optional[int] = None,
        save: bool = True,
    ):
        if x is not None and y is not None:
            self.setPos(QPoint(x, y))

    def paint(self, painter: QPainter, option, widget=None):
        painter.drawPixmap(self.pixmap.rect(), self.pixmap)

    def setPixmap(self, pixmap):
        self.pixmap = pixmap
        super().setPixmap(pixmap)



from PySide6.QtCore import Qt, QSize
from PySide6.QtGui import (
    QStandardItemModel,
    QStandardItem,
    QBrush,
    QColor,
    QPainter,
)
from PySide6.QtWidgets import (
    QStyledItemDelegate,
    QStyleOptionViewItem,
    QLabel,
    QWidget,
    QVBoxLayout,
    QPlainTextEdit,
)

from airunner.enums import SignalCode
from airunner.gui.widgets.api_token.api_token_widget import APITokenWidget
from airunner.gui.widgets.export_preferences.export_preferences_widget import (
    ExportPreferencesWidget,
)
from airunner.gui.widgets.keyboard_shortcuts.keyboard_shortcuts_widget import (
    KeyboardShortcutsWidget,
)
from airunner.gui.widgets.llm.bot_preferences import BotPreferencesWidget
from airunner.gui.widgets.llm.prompt_templates_widget import (
    PromptTemplatesWidget,
)
from airunner.gui.widgets.memory_preferences.memory_preferences_widget import (
    MemoryPreferencesWidget,
)
from airunner.gui.widgets.paths.paths_widget import PathsWidget
from airunner.gui.widgets.tts.voice_settings_widget import VoiceSettingsWidget
from airunner.gui.widgets.user.user_settings_widget import UserSettingsWidget
from airunner.gui.widgets.sound_settings.sound_settings_widget import (
    SoundSettingsWidget,
)
from airunner.gui.widgets.huggingface_settings.huggingface_settings_widget import (
    HuggingfaceSettingsWidget,
)
from airunner.gui.widgets.rag_settings.rag_settings_widget import (
    RAGSettingsWidget,
)
from airunner.gui.widgets.openrouter_settings.openrouter_settings_widget import (
    OpenrouterSettingsWidget,
)
from airunner.gui.windows.settings.templates.airunner_settings_ui import (
    Ui_airunner_settings,
)
from airunner.gui.windows.base_window import BaseWindow
from airunner.utils.settings import get_qsettings


class HighlightDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.selected_indexes = []

    def paint(self, painter: QPainter, option: QStyleOptionViewItem, index):
        if index in self.selected_indexes:
            painter.save()
            painter.fillRect(
                option.rect, QBrush(QColor(0, 0, 255, 50))
            )  # Blue background color
            painter.restore()

        super().paint(painter, option, index)

    def add_selected_index(self, index):
        self.selected_indexes.append(index)

    def remove_selected_index(self, index):
        self.selected_indexes.remove(index)


class SettingsWindow(BaseWindow):
    is_modal = True
    template_class_ = Ui_airunner_settings

    def __init__(self, **kwargs):
        self.widgets = {}
        self.qsettings = get_qsettings()
        super().__init__(**kwargs)
        self.model = None
        self.scroll_widget = None
        self.scroll_layout = None
        self.highlight_delegate = None
        self.emit_signal(SignalCode.APPLICATION_SETTINGS_LOADED_SIGNAL)

    def showEvent(self, event):
        super().showEvent(event)
        self.setWindowTitle("AI Runner Preferences")

    @staticmethod
    def available_widgets(name):
        if name == "paths":
            return PathsWidget
        elif name == "keyboard_shortcuts":
            return KeyboardShortcutsWidget
        elif name == "memory":
            return MemoryPreferencesWidget
        elif name == "hf_api_key":
            return APITokenWidget
        elif name == "voice_settings":
            return VoiceSettingsWidget
        elif name == "prompt_templates":
            return PromptTemplatesWidget
        elif name == "bot_preferences":
            return BotPreferencesWidget
        elif name == "user_settings":
            return UserSettingsWidget
        elif name == "export_preferences":
            return ExportPreferencesWidget
        elif name == "sound_settings":
            return SoundSettingsWidget
        elif name == "huggingface_settings":
            return HuggingfaceSettingsWidget
        elif name == "rag_settings":
            return RAGSettingsWidget
        elif name == "openrouter_settings":
            return OpenrouterSettingsWidget
        # elif name == "stt_preferences":
        #     return STTSettingsWidget

    def get_callback_for_slider(self, callback_name):
        return getattr(self, callback_name)

    def initialize_window(self):
        self.model = QStandardItemModel()
        self.ui.directory.setModel(self.model)
        self.ui.directory.setHeaderHidden(True)
        self.ui.directory.setIndentation(20)

        directory = [
            {
                "section": "Image Export Preferences",
                "files": [
                    {
                        "name": "export_preferences",
                        "display_name": "Export Preferences",
                        "checkable": False,
                    }
                ],
            },
            {
                "section": "Memory Preferences",
                "files": [
                    {
                        "name": "memory",
                        "display_name": "Memory",
                        "checkable": False,
                    }
                ],
            },
            {
                "section": "Keyboard Shortcuts",
                "files": [
                    {
                        "name": "keyboard_shortcuts",
                        "display_name": "Keyboard Shortcuts",
                        "checkable": False,
                    }
                ],
            },
            {
                "section": "Chatbot, prompt templates, text-to-speech",
                "files": [
                    {
                        "name": "bot_preferences",
                        "display_name": "Agent Preferences",
                        "checkable": False,
                    },
                    {
                        "name": "user_settings",
                        "display_name": "User Settings",
                        "checkable": False,
                    },
                    {
                        "name": "voice_settings",
                        "display_name": "Text-to-Speech",
                        "checkable": False,
                    },
                    {
                        "name": "prompt_templates",
                        "display_name": "Prompt Templates",
                        "checkable": False,
                    },
                    # {
                    #     "name": "stt_preferences",
                    #     "display_name": "Speech-to-Text",
                    #     "checkable": False
                    # },
                ],
            },
            {
                "section": "Miscellaneous Preferences",
                "files": [
                    {
                        "name": "dark_mode",
                        "display_name": "Dark Mode",
                        "checkable": True,
                        "description": "If enabled, AI Runner will use a dark theme.",
                    },
                    {
                        "name": "override_system_theme",
                        "display_name": "Override System Theme",
                        "checkable": True,
                        "description": "If enabled, override the system theme with the selected theme.",
                    },
                    {
                        "name": "check_for_updates",
                        "display_name": "Check for updates",
                        "checkable": True,
                        "description": "If enabled, AI Runner will check for updates on startup.",
                    },
                    {
                        "name": "enable_workflows",
                        "display_name": "Enable Workflows",
                        "checkable": True,
                        "description": "Enable experimental agent workflows.",
                    },
                    {
                        "name": "sound_settings",
                        "display_name": "Sound Settings",
                        "checkable": False,
                    },
                    {
                        "name": "huggingface_settings",
                        "display_name": "Huggingface Settings",
                        "checkable": False,
                    },
                    {
                        "name": "rag_settings",
                        "display_name": "RAG Settings",
                        "checkable": False,
                    },
                    {
                        "name": "openrouter_settings",
                        "display_name": "Openrouter Settings",
                        "checkable": False,
                    },
                ],
            },
        ]
        for index, section in enumerate(directory):
            self.add_directory(section["section"])
            for file in section["files"]:
                self.add_file(
                    self.model.item(index),
                    file["name"],
                    file["display_name"],
                    file["checkable"],
                    file["description"] if "description" in file else None,
                )

        self.scroll_widget = QWidget()
        self.scroll_layout = QVBoxLayout(self.scroll_widget)
        self.ui.preferences.setWidget(self.scroll_widget)

        self.highlight_delegate = HighlightDelegate(self.ui.directory)
        self.ui.directory.setItemDelegate(self.highlight_delegate)

        self.ui.directory.clicked.connect(self.on_item_clicked)

        # expand all directories
        for i in range(self.model.rowCount()):
            index = self.model.index(i, 0)
            self.ui.directory.setExpanded(index, True)

        self.connect_signals()

    def connect_signals(self):
        self.register(SignalCode.VOICE_SAVED, self.refresh_bot_preferences)

    def refresh_bot_preferences(self):
        if "bot_preferences" in self.widgets:
            self.widgets["bot_preferences"].load_voices()

    def add_directory(self, name, checkable=False):
        item = QStandardItem(name)
        item.setCheckable(checkable)
        if checkable:
            checked = False
            if name == "enable_workflows":
                checked = self.enable_workflows
            item.setCheckState(
                Qt.CheckState.Checked if checked else Qt.CheckState.Unchecked
            )
        item.setFlags(item.flags() & ~Qt.ItemFlag.ItemIsEditable)
        item.setData(name, Qt.ItemDataRole.DisplayRole)
        item.setSizeHint(QSize(0, 24))
        self.model.appendRow(item)

    def add_file(
        self,
        parent_item,
        name,
        display_name,
        checkable=False,
        description=None,
    ):
        file_item = QStandardItem(display_name)
        file_item.setCheckable(checkable)
        if checkable:
            checked = False
            if name == "resize_on_import":
                checked = self.application_settings.resize_on_paste
            elif name == "image_to_new_layer":
                checked = self.application_settings.image_to_new_layer is True
            elif name == "dark_mode":
                checked = self.application_settings.dark_mode_enabled
            elif name == "override_system_theme":
                checked = self.application_settings.override_system_theme
            elif name == "check_for_updates":
                checked = self.application_settings.latest_version_check
            elif name == "enable_workflows":
                checked = self.enable_workflows
            elif name == "allow_online_mode":
                checked = self.application_settings.allow_online_mode

            file_item.setCheckState(
                Qt.CheckState.Checked if checked else Qt.CheckState.Unchecked
            )
        # prevent file_item from being edited
        file_item.setFlags(file_item.flags() & ~Qt.ItemFlag.ItemIsEditable)
        file_item.setData(name, Qt.ItemDataRole.UserRole)
        file_item.setData(display_name, Qt.ItemDataRole.DisplayRole)
        file_item.setData(description, Qt.ItemDataRole.ToolTipRole)
        file_item.setSizeHint(QSize(0, 24))
        parent_item.appendRow(file_item)

    @property
    def enable_workflows(self) -> bool:
        return self.qsettings.value("enable_workflows") == "true"

    @enable_workflows.setter
    def enable_workflows(self, val: bool):
        self.qsettings.setValue("enable_workflows", val)
        self.emit_signal(SignalCode.ENABLE_WORKFLOWS_TOGGLED, {"enabled": val})
        self.qsettings.sync()

    def on_item_clicked(self, index):
        item = self.model.itemFromIndex(index)
        if item.parent() is None:
            return

        section = item.parent().data(Qt.ItemDataRole.DisplayRole)
        name = item.data(Qt.ItemDataRole.UserRole)
        display_name = item.data(Qt.ItemDataRole.DisplayRole)
        description = item.data(Qt.ItemDataRole.ToolTipRole)

        if name == "resize_on_import":
            checked = item.checkState() == Qt.CheckState.Checked
            self.update_application_settings("resize_on_paste", checked)
        elif name == "image_to_new_layer":
            checked = item.checkState() == Qt.CheckState.Checked
            self.update_application_settings("image_to_new_layer", checked)
        elif name == "dark_mode":
            checked = item.checkState() == Qt.CheckState.Checked
            self.update_application_settings("dark_mode_enabled", checked)
            self.emit_signal(SignalCode.REFRESH_STYLESHEET_SIGNAL)
        elif name == "override_system_theme":
            checked = item.checkState() == Qt.CheckState.Checked
            self.update_application_settings("override_system_theme", checked)
            self.emit_signal(SignalCode.REFRESH_STYLESHEET_SIGNAL)
        elif name == "check_for_updates":
            checked = item.checkState() == Qt.CheckState.Checked
            self.update_application_settings("latest_version_check", checked)
        elif name == "enable_workflows":
            checked = item.checkState() == Qt.CheckState.Checked
            self.enable_workflows = checked
        elif name == "allow_online_mode":
            checked = item.checkState() == Qt.CheckState.Checked
            self.update_application_settings("allow_online_mode", checked)
        elif name == "reset_settings":
            self.emit_signal(SignalCode.APPLICATION_RESET_SETTINGS_SIGNAL)

        self.show_content(section, display_name, name, description)
        self.set_stylesheet()

    def show_content(self, section, display_name, name, description):
        # create a label widget
        label = QLabel(f"<p><b>{section} > {display_name}</b></p>")

        widget_object = None
        if widget_object is None:
            _widget_class = self.available_widgets(name)
            if _widget_class is not None:
                widget_object = _widget_class()
        if widget_object is None:
            return

        self.clear_scroll_area()
        self.scroll_layout.addWidget(label)
        if widget_object:
            self.scroll_layout.addWidget(widget_object)
        elif description:
            description_text_edit = QPlainTextEdit()
            description_text_edit.setReadOnly(True)
            description_text_edit.setPlainText(description)
            description_text_edit.setFrameStyle(0)
            description_text_edit.setStyleSheet(
                "QPlainTextEdit { background-color: transparent; border: none; }"
            )
            self.scroll_layout.addWidget(description_text_edit)

    def clear_scroll_area(self):
        while self.scroll_layout.count():
            item = self.scroll_layout.takeAt(0)
            widget = item.widget()
            if widget:
                widget.deleteLater()



import unittest
from unittest.mock import patch, MagicMock

from sqlalchemy.orm import Session

from airunner.data.session_manager import session_scope, Session as SessionClass, engine


class TestSessionManager(unittest.TestCase):
    """Test cases for database session management."""

    def test_session_class_exists(self):
        """Test that the Session class is properly configured."""
        self.assertIsNotNone(SessionClass)
        # Check that Session is callable (creates a session)
        session = SessionClass()
        self.assertIsInstance(session, Session)
        session.close()

    def test_engine_exists(self):
        """Test that the database engine is properly configured."""
        self.assertIsNotNone(engine)
        # Verify engine has expected attributes
        self.assertTrue(hasattr(engine, 'connect'))
        self.assertTrue(hasattr(engine, 'dispose'))
        
    @patch('airunner.data.session_manager.Session')
    def test_session_scope_normal_execution(self, mock_session):
        """Test the session_scope context manager with normal execution."""
        # Set up mocks
        mock_session_instance = MagicMock()
        mock_session.return_value = mock_session_instance
        
        # Use the context manager
        with session_scope() as session:
            # Verify session is the mock and that we can use it
            self.assertEqual(session, mock_session_instance)
            session.query()  # Do something with the session
        
        # Verify the expected methods were called
        mock_session.assert_called_once()
        mock_session_instance.commit.assert_called_once()
        mock_session_instance.close.assert_called_once()
        mock_session_instance.rollback.assert_not_called()

    @patch('airunner.data.session_manager.Session')
    def test_session_scope_with_exception(self, mock_session):
        """Test the session_scope context manager when an exception occurs."""
        # Set up mocks
        mock_session_instance = MagicMock()
        mock_session.return_value = mock_session_instance
        
        # Use the context manager with an exception
        with self.assertRaises(ValueError):
            with session_scope() as session:
                # Do something that raises an exception
                raise ValueError("Test exception")
        
        # Verify the expected methods were called
        mock_session.assert_called_once()
        mock_session_instance.commit.assert_not_called()
        mock_session_instance.rollback.assert_called_once()
        mock_session_instance.close.assert_called_once()

    @patch('airunner.data.session_manager.Session')
    def test_session_scope_cleanup_on_exception_in_commit(self, mock_session):
        """Test that session is cleaned up even when commit fails."""
        # Set up mocks
        mock_session_instance = MagicMock()
        mock_session_instance.commit.side_effect = Exception("Commit failed")
        mock_session.return_value = mock_session_instance
        
        # Use the context manager expecting an exception
        with self.assertRaises(Exception):
            with session_scope() as session:
                pass  # Normal execution, but commit will fail
        
        # Verify the expected methods were called
        mock_session.assert_called_once()
        mock_session_instance.commit.assert_called_once()
        mock_session_instance.rollback.assert_called_once()
        mock_session_instance.close.assert_called_once()


if __name__ == "__main__":
    unittest.main()


__all__ = [
    "create_worker",
    "get_lat_lon",
    "get_torch_device",
    "get_version",
    "open_file_path",
    "parse_template",
    "get_platform_name",
    "is_linux",
    "is_bsd",
    "is_darwin",
    "is_windows",
    "random_seed",
    "set_widget_state",
    "snap_to_grid",
    "strip_names_from_message",
    "prepare_text_for_tts",
    "replace_unspeakable_characters",
    "strip_emoji_characters",
    "replace_numbers_with_words",
    "replace_misc_with_words",
    "roman_to_int",
]


def __getattr__(name):
    if name == "create_worker":
        from .application.create_worker import create_worker

        return create_worker
    elif name == "get_lat_lon":
        from .location.get_lat_lon import get_lat_lon

        return get_lat_lon
    elif name == "get_torch_device":
        from .application.get_torch_device import get_torch_device

        return get_torch_device
    elif name == "get_version":
        from .application.get_version import get_version

        return get_version
    elif name == "open_file_path":
        from .os.open_file_path import open_file_path

        return open_file_path
    elif name == "parse_template":
        from .llm.parse_template import parse_template

        return parse_template
    elif name == "get_platform_name":
        from .application.platform_info import get_platform_name

        return get_platform_name
    elif name == "is_linux":
        from .application.platform_info import is_linux

        return is_linux
    elif name == "is_bsd":
        from .application.platform_info import is_bsd

        return is_bsd
    elif name == "is_darwin":
        from .application.platform_info import is_darwin

        return is_darwin
    elif name == "is_windows":
        from .application.platform_info import is_windows

        return is_windows
    elif name == "random_seed":
        from .application.random_seed import random_seed

        return random_seed
    elif name == "set_widget_state":
        from .application.set_widget_state import set_widget_state

        return set_widget_state
    elif name == "snap_to_grid":
        from .application.snap_to_grid import snap_to_grid

        return snap_to_grid
    elif name == "strip_names_from_message":
        from .llm.strip_names_from_message import strip_names_from_message

        return strip_names_from_message
    elif name == "prepare_text_for_tts":
        from .llm.text_preprocessing import prepare_text_for_tts

        return prepare_text_for_tts
    elif name == "replace_unspeakable_characters":
        from .llm.text_preprocessing import replace_unspeakable_characters

        return replace_unspeakable_characters
    elif name == "strip_emoji_characters":
        from .llm.text_preprocessing import strip_emoji_characters

        return strip_emoji_characters
    elif name == "replace_numbers_with_words":
        from .llm.text_preprocessing import replace_numbers_with_words

        return replace_numbers_with_words
    elif name == "replace_misc_with_words":
        from .llm.text_preprocessing import replace_misc_with_words

        return replace_misc_with_words
    elif name == "roman_to_int":
        from .llm.text_preprocessing import roman_to_int

        return roman_to_int
    raise AttributeError(f"module {__name__} has no attribute {name}")



from sqlalchemy import Column, Integer, Boolean, LargeBinary

from airunner.data.models.base import BaseModel


class ImageToImageSettings(BaseModel):
    __tablename__ = 'image_to_image_settings'
    id = Column(Integer, primary_key=True, autoincrement=True)
    image = Column(LargeBinary, nullable=True)
    enabled = Column(Boolean, default=False)
    use_grid_image_as_input = Column(Boolean, default=False)
    lock_input_image = Column(Boolean, default=False)



import torch
from abc import ABC, abstractmethod, ABCMeta
from typing import Dict

"""
The following code ensures that we only use PySide6 if it is available.
If it is not available, we use a placeholder class instead.
"""
try:
    from PySide6.QtCore import QObject

    class OptionalQObject(QObject):
        pass

except ImportError:

    class OptionalQObject:
        """
        A placeholder class to avoid hard dependency on PySide6.
        """

        pass


from airunner.enums import (
    HandlerType,
    SignalCode,
    ModelType,
    ModelStatus,
    ModelAction,
)
from airunner.utils.application.mediator_mixin import MediatorMixin
from airunner.utils.application import get_torch_device
from airunner.gui.windows.main.settings_mixin import SettingsMixin
from airunner.settings import (
    AIRUNNER_MEM_LLM_DEVICE,
    AIRUNNER_MEM_SD_DEVICE,
    AIRUNNER_MEM_TTS_DEVICE,
    AIRUNNER_MEM_STT_DEVICE,
)

QObjectMeta = type(OptionalQObject)


class CombinedMeta(QObjectMeta, ABCMeta):
    pass


class BaseModelManager(
    MediatorMixin, SettingsMixin, OptionalQObject, ABC, metaclass=CombinedMeta
):
    """
    Base abstract class for all model handlers.
    Provides common functionality and interface for all handlers.
    """

    handler_type = HandlerType.TRANSFORMER
    model_type = None
    _model_status: Dict = {}

    def __init__(self, *args, **kwargs):
        self.use_gpu = True
        super().__init__()
        self._requested_action: ModelAction = ModelAction.NONE

    @property
    def model_status(self) -> Dict:
        return self._model_status

    @property
    def requested_action(self):
        return self._requested_action

    @requested_action.setter
    def requested_action(self, value):
        self._requested_action = value

    def handle_requested_action(self):
        if self._requested_action is ModelAction.LOAD:
            self.load()
            self._requested_action = ModelAction.NONE
        if self._requested_action is ModelAction.CLEAR:
            self.unload()
            self._requested_action = ModelAction.NONE

    @abstractmethod
    def load(self):
        """
        Load the model and related components.
        """

    @abstractmethod
    def unload(self):
        """
        Unload the model and free resources.
        """

    @property
    def device_index(self):
        device = None
        model_type_str = ""
        if self.model_type is ModelType.LLM:
            device = AIRUNNER_MEM_LLM_DEVICE
            model_type_str = "llm"
        elif self.model_type is ModelType.TTS:
            device = AIRUNNER_MEM_TTS_DEVICE
            model_type_str = "tts"
        elif self.model_type is ModelType.STT:
            device = AIRUNNER_MEM_STT_DEVICE
            model_type_str = "stt"
        elif self.model_type is ModelType.SD:
            device = AIRUNNER_MEM_SD_DEVICE
            model_type_str = "sd"
        if device is None:
            if not self.model_type:
                raise ValueError("model_type not set")
            device = getattr(
                self.memory_settings, f"default_gpu_{model_type_str}"
            )
        return device

    @property
    def device(self):
        return get_torch_device(self.device_index)

    @property
    def llm_dtype(self):
        return self.llm_generator_settings.dtype

    @property
    def use_cuda(self):
        if self.handler_type == HandlerType.TRANSFORMER and (
            self.llm_dtype == "32bit" or not self.use_gpu
        ):
            return False
        return self.application_settings.use_cuda and torch.cuda.is_available()

    @property
    def cuda_index(self):
        return 0

    @property
    def torch_dtype(self):
        return torch.float16 if self.use_cuda else torch.float32

    def change_model_status(self, model: ModelType, status: ModelStatus):
        self._model_status[model] = status
        self.emit_signal(
            SignalCode.MODEL_STATUS_CHANGED_SIGNAL,
            {"model": model, "status": status},
        )



from airunner.gui.widgets.nodegraph.nodes.llm.base_llm_node import (
    BaseLLMNode,
)


class AgentActionNode(BaseLLMNode):
    NODE_NAME = "Agent Action"

    def __init__(self):
        super().__init__()
        # Default ports - can be customized via properties or methods
        self.add_input("in_message")
        self.add_output("out_message")
        # Add a text input widget to specify the action class/name
        self.add_text_input("action_name", "Action Name", tab="widgets")

    # Override execute for AgentAction specific logic (placeholder)
    def execute(self, input_data):
        action_name = self.get_property("action_name")
        in_message = input_data.get(
            "in_message", None
        )  # Get data from the connected input port
        print(
            f"Executing Agent Action: {action_name} with input: {in_message}"
        )
        # Dummy logic: Find the corresponding AgentAction class based on action_name
        # and call its run method. For now, just pass data through.
        output_data = f"Action '{action_name}' processed: {in_message}"
        return {
            "out_message": output_data
        }  # Return data for the 'out_message' port



from PIL.ImageFilter import UnsharpMask as PILUnsharpMask

from airunner.filters.base_filter import BaseFilter


class UnsharpMask(BaseFilter):
    """Apply an unsharp mask filter to an image.
    
    This filter enhances edges and details by applying an unsharp mask.
    
    Attributes:
        radius: Controls the size of the edges to enhance.
        percent: Controls the amount of enhancement.
        threshold: Controls the minimum brightness change to be enhanced.
    """
    
    def apply_filter(self, image, do_reset=False):
        """Apply unsharp mask filter to the image.
        
        Args:
            image: The PIL Image to filter.
            do_reset: Whether to reset internal state (unused in this filter).
            
        Returns:
            The sharpened PIL Image.
        """
        return image.filter(
            PILUnsharpMask(
                radius=self.radius * 100,
                percent=int(self.percent * 200),
                threshold=int(self.threshold * 10)
            )
        )



# Random Number Generator Node
import random
from NodeGraphQt.constants import NodePropWidgetEnum
from airunner.gui.widgets.nodegraph.nodes.math.base_math_node import (
    BaseMathNode,
)


class RandomNumberNode(BaseMathNode):
    NODE_NAME = "Random Number Generator"

    def __init__(self):
        super().__init__()

        # Inputs for min and max values
        self.add_input("min")
        self.add_input("max")

        # Output for the random number
        self.add_output("random_value")

        # Add default min value using NodeGraphQt's built-in integer widget
        self.create_property(
            "default_min",
            0,
            widget_type=NodePropWidgetEnum.INT.value,
            range=(-1000, 1000),
            tab="settings",
        )

        # Add default max value using NodeGraphQt's built-in integer widget
        self.create_property(
            "default_max",
            100,
            widget_type=NodePropWidgetEnum.INT.value,
            range=(-1000, 1000),
            tab="settings",
        )

    def execute(self, input_data):
        # Get input values or use defaults from property widgets
        try:
            min_val = input_data.get("min")
            if min_val is None:
                min_val = int(self.get_property("default_min"))
            else:
                min_val = int(min_val)

            max_val = input_data.get("max")
            if max_val is None:
                max_val = int(self.get_property("default_max"))
            else:
                max_val = int(max_val)

            # Ensure min is less than max
            if min_val > max_val:
                min_val, max_val = max_val, min_val

            # Generate random number
            random_value = random.randint(min_val, max_val)

            return {"random_value": random_value}
        except (ValueError, TypeError) as e:
            print(f"Error in RandomNumberNode: {e}")
            # Return a default value if there's an error
            return {"random_value": 0}



from PIL import Image, ImageFilter

from airunner.filters.base_filter import BaseFilter


class PixelFilter(BaseFilter):
    """Apply a pixel art effect to an image.
    
    This filter creates a pixel art effect by reducing colors and
    resizing the image with nearest-neighbor sampling.
    
    Attributes:
        number_of_colors: Number of colors to use in the quantized image.
        base_size: Base size for downscaling step.
        smoothing: Amount of smoothing to apply (0 for none).
    """
    
    def __init__(self, **kwargs):
        """Initialize the pixel art filter.
        
        Args:
            **kwargs: Arbitrary keyword arguments that will be set as
                     attributes on the filter instance.
        """
        super().__init__(**kwargs)
        self.current_number_of_colors = 0
    
    def apply_filter(self, image, do_reset=False):
        """Apply pixel art filter to the image.
        
        Args:
            image: The PIL Image to filter.
            do_reset: Whether to reset internal state.
            
        Returns:
            The pixel art PIL Image.
        """
        # Retrieve filter parameters with defaults
        number_of_colors = getattr(self, "number_of_colors", 24)
        base_size = getattr(self, "base_size", 256)
        smoothing = getattr(self, "smoothing", 0)
        
        # Ensure number_of_colors is an integer divisible by 2
        number_of_colors = int(number_of_colors) // 2 * 2
        
        # Quantize the image (reduce colors) if needed
        if self.current_number_of_colors != number_of_colors or do_reset:
            try:
                self.current_number_of_colors = number_of_colors
                quantized = image.quantize(number_of_colors)
                self.image = quantized.convert("RGBA")
            except ValueError:
                self.logger.debug("Bad number of colors")
        
        image = self.image
        
        # Downsize while maintaining aspect ratio
        width, height = image.size
        scale = min(base_size / width, base_size / height)
        new_width = int(width * scale)
        new_height = int(height * scale)
        downsized = image.resize(
            (new_width, new_height), 
            Image.Resampling.NEAREST
        )
        
        # Upscale back to original dimensions
        target_width = int(new_width / scale)
        target_height = int(new_height / scale)
        final_image = downsized.resize(
            (target_width, target_height), 
            Image.Resampling.NEAREST
        )
        
        # Apply smoothing if enabled
        if smoothing > 0:
            for _ in range(smoothing // 10):  # Apply smoothing filter multiple times
                final_image = final_image.filter(ImageFilter.SMOOTH)
        
        return final_image



from typing import Any, Callable, List, Optional, Sequence, Union, Dict
import torch
from llama_index.core.base.llms.types import (
    ChatMessage,
    ChatResponse,
    ChatResponseGen,
    CompletionResponse,
    CompletionResponseGen,
    LLMMetadata,
)
from llama_index.core.bridge.pydantic import Field, PrivateAttr
from llama_index.core.callbacks import CallbackManager
from llama_index.core.constants import (
    DEFAULT_CONTEXT_WINDOW,
    DEFAULT_NUM_OUTPUTS,
)
from llama_index.core.llms.callbacks import (
    llm_chat_callback,
    llm_completion_callback,
)
from llama_index.core.llms.custom import CustomLLM
from llama_index.core.base.llms.generic_utils import (
    completion_response_to_chat_response,
    stream_completion_response_to_chat_response,
    messages_to_prompt as generic_messages_to_prompt,
)
from llama_index.core.prompts.base import PromptTemplate
from llama_index.core.types import (
    BaseOutputParser,
    PydanticProgramMode,
    Thread,
)
from transformers import (
    AutoModelForCausalLM,
    AutoTokenizer,
    StoppingCriteria,
    StoppingCriteriaList,
)

from airunner.gui.windows.main.settings_mixin import SettingsMixin
from airunner.settings import AIRUNNER_DEFAULT_LLM_HF_PATH
from airunner.handlers.llm.llm_request import LLMRequest


class HuggingFaceLLM(CustomLLM, SettingsMixin):
    """
    HuggingFace LLM integration for AI Runner.

    This class provides an interface between AI Runner and HuggingFace models,
    allowing for text completion and chat functionality with various models.
    """

    model_name: str = Field(
        default=AIRUNNER_DEFAULT_LLM_HF_PATH,
        description="The model name to use from HuggingFace. Unused if `model` is passed directly.",
    )
    model: Optional[Any] = Field(
        default=None,
        description="The model to use. If not passed in, will be loaded from `model_name`.",
    )
    context_window: int = Field(
        default=DEFAULT_CONTEXT_WINDOW,
        description="The maximum number of tokens available for input.",
        gt=0,
    )
    max_new_tokens: int = Field(
        default=DEFAULT_NUM_OUTPUTS,
        description="The maximum number of tokens to generate.",
        gt=0,
    )
    system_prompt: str = Field(
        default="",
        description="The system prompt with extra instructions or context.",
    )
    query_wrapper_prompt: PromptTemplate = Field(
        default=PromptTemplate("{query_str}"),
        description="The query wrapper prompt containing the query placeholder.",
    )
    tokenizer_name: str = Field(
        default=AIRUNNER_DEFAULT_LLM_HF_PATH,
        description="The name of the tokenizer to use from HuggingFace.",
    )
    device_map: str = Field(
        default="auto",
        description="The device_map to use. Defaults to 'auto'.",
    )
    stopping_ids: List[int] = Field(
        default_factory=list,
        description="Token IDs that signal generation should stop when predicted.",
    )
    tokenizer_outputs_to_remove: list = Field(
        default_factory=list,
        description="Outputs to remove from the tokenizer to prevent errors.",
    )
    tokenizer_kwargs: Dict = Field(
        default_factory=dict,
        description="The kwargs to pass to the tokenizer.",
    )
    model_kwargs: Dict = Field(
        default_factory=dict,
        description="The kwargs to pass to the model during initialization.",
    )

    _llm_request: Optional[LLMRequest] = None

    @property
    def generate_kwargs(self) -> Dict:
        """
        Get generation parameters from the LLM request.

        Returns:
            Dict: Dictionary of generation parameters.
        """
        if self.llm_request:
            data = self.llm_request.to_dict()

            # Remove unused keys
            del data["do_tts_reply"]
            return data
        return {}

    @property
    def llm_request(self) -> Optional[LLMRequest]:
        """
        Get the current LLM request, initializing with defaults if needed.

        Returns:
            Optional[LLMRequest]: The current LLM request configuration.
        """
        if self._llm_request is None:
            self._llm_request = LLMRequest.from_default()
        return self._llm_request

    @llm_request.setter
    def llm_request(self, value: Optional[LLMRequest]):
        """
        Set the LLM request configuration.

        Args:
            value: The new LLM request configuration.
        """
        self._llm_request = value

    is_chat_model: bool = Field(
        default=False,
        description="Whether the model is designed for chat interactions.",
    )

    _model: Any = PrivateAttr()
    _tokenizer: Any = PrivateAttr()
    _stopping_criteria: Any = PrivateAttr()
    _streaming_stopping_criteria: Any = PrivateAttr(default=None)

    def __init__(
        self,
        context_window: int = DEFAULT_CONTEXT_WINDOW,
        max_new_tokens: int = DEFAULT_NUM_OUTPUTS,
        query_wrapper_prompt: Union[str, PromptTemplate] = "{query_str}",
        tokenizer_name: str = AIRUNNER_DEFAULT_LLM_HF_PATH,
        model_name: str = AIRUNNER_DEFAULT_LLM_HF_PATH,
        model: Optional[Any] = None,
        tokenizer: Optional[Any] = None,
        device_map: Optional[str] = "auto",
        stopping_ids: Optional[List[int]] = None,
        tokenizer_kwargs: Optional[Dict] = None,
        tokenizer_outputs_to_remove: Optional[list] = None,
        model_kwargs: Optional[Dict] = None,
        generate_kwargs: Optional[Dict] = None,
        is_chat_model: Optional[bool] = False,
        callback_manager: Optional[CallbackManager] = None,
        system_prompt: str = "",
        messages_to_prompt: Optional[
            Callable[[Sequence[ChatMessage]], str]
        ] = None,
        completion_to_prompt: Optional[Callable[[str], str]] = None,
        pydantic_program_mode: PydanticProgramMode = PydanticProgramMode.DEFAULT,
        output_parser: Optional[BaseOutputParser] = None,
        streaming_stopping_criteria: Optional[StoppingCriteria] = None,
    ) -> None:
        """
        Initialize the HuggingFace LLM.

        Args:
            context_window: Maximum tokens for input context.
            max_new_tokens: Maximum tokens to generate.
            query_wrapper_prompt: Template for wrapping queries.
            tokenizer_name: Name of the HuggingFace tokenizer.
            model_name: Name of the HuggingFace model.
            model: Pre-loaded model instance (optional).
            tokenizer: Pre-loaded tokenizer instance (optional).
            device_map: Device mapping strategy.
            stopping_ids: Token IDs that signal generation should stop.
            tokenizer_kwargs: Additional kwargs for tokenizer initialization.
            tokenizer_outputs_to_remove: Tokenizer outputs to remove.
            model_kwargs: Additional kwargs for model initialization.
            generate_kwargs: Additional kwargs for generation.
            is_chat_model: Whether the model is designed for chat interactions.
            callback_manager: Manager for callbacks.
            system_prompt: System prompt for additional context.
            messages_to_prompt: Function to convert chat messages to prompt.
            completion_to_prompt: Function to convert completion to prompt.
            pydantic_program_mode: Mode for Pydantic program generation.
            output_parser: Parser for model outputs.
            streaming_stopping_criteria: Criteria for stopping streaming.
        """
        model_kwargs = model_kwargs or {}
        model = model or AutoModelForCausalLM.from_pretrained(
            model_name, device_map=device_map, **model_kwargs
        )

        # Check context_window against model's capabilities
        config_dict = model.config.to_dict()
        model_context_window = int(
            config_dict.get("max_position_embeddings", context_window)
        )
        if model_context_window and model_context_window < context_window:
            self.logger.warning(
                f"Supplied context_window {context_window} is greater "
                f"than the model's max input size {model_context_window}. "
                "Using model's max input size instead."
            )
            context_window = model_context_window

        # Initialize tokenizer
        tokenizer_kwargs = tokenizer_kwargs or {}
        tokenizer = tokenizer or AutoTokenizer.from_pretrained(
            tokenizer_name, **tokenizer_kwargs
        )

        # Set up stopping criteria
        stopping_ids_list = stopping_ids or []

        class StopOnTokens(StoppingCriteria):
            """Stopping criteria that stops generation when specific tokens are generated."""

            def __call__(
                self,
                input_ids: torch.LongTensor,
                scores: torch.FloatTensor,
                **kwargs: Any,
            ) -> bool:
                for stop_id in stopping_ids_list:
                    if input_ids[0][-1] == stop_id:
                        return True
                return False

        stopping_criteria = StoppingCriteriaList([StopOnTokens()])

        # Convert string prompt template to PromptTemplate if needed
        if isinstance(query_wrapper_prompt, str):
            query_wrapper_prompt = PromptTemplate(query_wrapper_prompt)

        # Use tokenizer-specific method or fallback to generic for messages_to_prompt
        messages_to_prompt = (
            messages_to_prompt or self._tokenizer_messages_to_prompt
        )

        super().__init__(
            context_window=context_window,
            max_new_tokens=max_new_tokens,
            query_wrapper_prompt=query_wrapper_prompt,
            tokenizer_name=tokenizer_name,
            model_name=model_name,
            device_map=device_map,
            stopping_ids=stopping_ids or [],
            tokenizer_kwargs=tokenizer_kwargs or {},
            tokenizer_outputs_to_remove=tokenizer_outputs_to_remove or [],
            model_kwargs=model_kwargs or {},
            generate_kwargs=generate_kwargs or {},
            is_chat_model=is_chat_model,
            callback_manager=callback_manager,
            system_prompt=system_prompt,
            messages_to_prompt=messages_to_prompt,
            completion_to_prompt=completion_to_prompt,
            pydantic_program_mode=pydantic_program_mode,
            output_parser=output_parser,
        )

        self._model = model
        self._tokenizer = tokenizer
        self._stopping_criteria = stopping_criteria

        # Set up streaming stopping criteria if provided
        if streaming_stopping_criteria:
            self._streaming_stopping_criteria = StoppingCriteriaList(
                [streaming_stopping_criteria, StopOnTokens()]
            )
        else:
            self._streaming_stopping_criteria = stopping_criteria

    @property
    def model(self):
        """Get the underlying HuggingFace model."""
        return self._model

    @property
    def tokenizer(self):
        """Get the underlying HuggingFace tokenizer."""
        return self._tokenizer

    @classmethod
    def class_name(cls) -> str:
        """Get the class name for registration."""
        return "HuggingFace_LLM"

    @property
    def metadata(self) -> LLMMetadata:
        """
        Get LLM metadata.

        Returns:
            LLMMetadata: Metadata about the model capabilities.
        """
        return LLMMetadata(
            context_window=self.context_window,
            num_output=self.max_new_tokens,
            model_name=self.model_name,
            is_chat_model=self.is_chat_model,
            is_function_calling_model=True,
        )

    def unload(self):
        """
        Unload the model and tokenizer from memory.

        This method helps with resource management by releasing GPU memory.
        """
        self.logger.debug("Unloading HuggingFace LLM")
        del self._model
        del self._tokenizer
        del self._stopping_criteria
        self._model = None
        self._tokenizer = None
        self._stopping_criteria = None

    def _tokenizer_messages_to_prompt(
        self, messages: Sequence[ChatMessage]
    ) -> str:
        """
        Convert chat messages to a prompt string using the tokenizer.

        Uses the tokenizer's chat template if available, otherwise falls back
        to generic conversion.

        Args:
            messages: Sequence of chat messages to convert.

        Returns:
            str: Prompt string suitable for the model.
        """
        if hasattr(self._tokenizer, "apply_chat_template"):
            messages_dict = [
                {"role": message.role.value, "content": message.content}
                for message in messages
            ]
            return self._tokenizer.apply_chat_template(
                messages_dict, tokenize=False, add_generation_prompt=True
            )
        return generic_messages_to_prompt(messages)

    @llm_completion_callback()
    def complete(
        self, prompt: str, formatted: bool = False, **kwargs: Any
    ) -> CompletionResponse:
        """
        Generate a completion for the given prompt.

        Args:
            prompt: The input prompt for text generation.
            formatted: Whether the prompt is already formatted.
            **kwargs: Additional keyword arguments for completion.

        Returns:
            CompletionResponse: The model's generated response.
        """
        full_prompt = prompt
        if not formatted:
            if self.query_wrapper_prompt:
                full_prompt = self.query_wrapper_prompt.format(
                    query_str=prompt
                )
            if self.completion_to_prompt:
                full_prompt = self.completion_to_prompt(full_prompt)
            elif self.system_prompt:
                full_prompt = f"{self.system_prompt} {full_prompt}"

        # Tokenize the prompt
        inputs = self._tokenizer(full_prompt, return_tensors="pt")
        inputs = inputs.to(self._model.device)

        # Remove keys from the tokenizer if needed, to avoid HF errors
        for key in self.tokenizer_outputs_to_remove:
            if key in inputs:
                inputs.pop(key, None)

        # Generate response
        tokens = self._model.generate(
            **inputs,
            stopping_criteria=self._stopping_criteria,
            **self.generate_kwargs,
        )

        # Extract and decode the generated text
        completion_tokens = tokens[0][inputs["input_ids"].size(1) :]
        completion = self._tokenizer.decode(
            completion_tokens, skip_special_tokens=True
        )

        return CompletionResponse(
            text=completion, raw={"model_output": tokens}
        )

    @llm_completion_callback()
    def stream_complete(
        self, prompt: str, formatted: bool = False, **kwargs: Any
    ) -> CompletionResponseGen:
        """
        Stream a completion for the given prompt.

        Args:
            prompt: The input prompt for text generation.
            formatted: Whether the prompt is already formatted.
            **kwargs: Additional keyword arguments for completion.

        Returns:
            CompletionResponseGen: Generator yielding parts of the response as they're generated.
        """
        from transformers import TextIteratorStreamer

        full_prompt = prompt
        if not formatted:
            if self.query_wrapper_prompt:
                full_prompt = self.query_wrapper_prompt.format(
                    query_str=prompt
                )
            if self.system_prompt:
                full_prompt = f"{self.system_prompt} {full_prompt}"

        # Tokenize the prompt
        inputs = self._tokenizer(full_prompt, return_tensors="pt")
        inputs = inputs.to(self._model.device)

        # Remove keys from the tokenizer if needed, to avoid HF errors
        for key in self.tokenizer_outputs_to_remove:
            if key in inputs:
                inputs.pop(key, None)

        # Set up streaming
        streamer = TextIteratorStreamer(
            self._tokenizer, skip_prompt=True, skip_special_tokens=True
        )

        generation_kwargs = dict(
            inputs,
            stopping_criteria=self._streaming_stopping_criteria,
            **self.generate_kwargs,
        )
        generation_kwargs["streamer"] = streamer

        # Generate in background thread
        thread = Thread(target=self._model.generate, kwargs=generation_kwargs)
        thread.start()

        # Create generator based on streamer
        def gen() -> CompletionResponseGen:
            text = ""
            for x in streamer:
                text += x
                yield CompletionResponse(text=text, delta=x)

        return gen()

    @llm_chat_callback()
    def chat(
        self, messages: Sequence[ChatMessage], **kwargs: Any
    ) -> ChatResponse:
        """
        Generate a chat response for the given messages.

        Args:
            messages: Sequence of chat messages for the conversation.
            **kwargs: Additional keyword arguments for chat.

        Returns:
            ChatResponse: The model's generated chat response.
        """
        prompt = self.messages_to_prompt(messages)
        completion_response = self.complete(prompt, formatted=True, **kwargs)
        return completion_response_to_chat_response(completion_response)

    @llm_chat_callback()
    def stream_chat(
        self, messages: Sequence[ChatMessage], **kwargs: Any
    ) -> ChatResponseGen:
        """
        Stream a chat response for the given messages.

        Args:
            messages: Sequence of chat messages for the conversation.
            **kwargs: Additional keyword arguments for chat.

        Returns:
            ChatResponseGen: Generator yielding parts of the chat response as they're generated.
        """
        prompt = self.messages_to_prompt(messages)
        completion_response = self.stream_complete(
            prompt, formatted=True, **kwargs
        )
        return stream_completion_response_to_chat_response(completion_response)



import sounddevice as sd
from typing import Optional
from queue import Queue

from PySide6.QtCore import QThread

from airunner.enums import SignalCode, TTSModel
from airunner.settings import AIRUNNER_SLEEP_TIME_IN_MS
from airunner.workers.worker import Worker


class TTSVocalizerWorker(Worker):
    """
    Speech (in the form of numpy arrays generated with the TTS class) is added to the
    vocalizer's queue. The vocalizer plays the speech using sounddevice.
    """

    reader_mode_active = False

    def __init__(self):
        self.signal_handlers = {
            SignalCode.INTERRUPT_PROCESS_SIGNAL: self.on_interrupt_process_signal,
            SignalCode.UNBLOCK_TTS_GENERATOR_SIGNAL: self.on_unblock_tts_generator_signal,
            SignalCode.TTS_GENERATOR_WORKER_ADD_TO_STREAM_SIGNAL: self.on_tts_generator_worker_add_to_stream_signal,
            SignalCode.APPLICATION_SETTINGS_CHANGED_SIGNAL: self.on_application_settings_changed_signal,
            SignalCode.PLAYBACK_DEVICE_CHANGED: self.on_playback_device_changed_signal,
        }
        self._selected_device = None
        super().__init__()
        self.queue = Queue()
        # check if speakers are available
        self.stream = None
        # self.start_stream()
        self.started = False
        self.do_interrupt = False
        self.accept_message = True

    @property
    def is_espeak(self) -> bool:
        return self.chatbot_voice_model_type == TTSModel.ESPEAK.value

    def on_interrupt_process_signal(self):
        self.stream.abort()
        self.accept_message = False
        self.queue = Queue()

    def on_unblock_tts_generator_signal(self):
        if self.application_settings.tts_enabled:
            self.logger.debug("Starting TTS stream...")
            self.accept_message = True
            self.stream.start()

    def on_application_settings_changed_signal(self, data):
        if (
            data
            and data.get("setting_name", "") == "speech_t5_settings"
            and data.get("column_name", "") == "pitch"
        ):
            pitch = data.get("value", 0)
            if self.stream is not None:
                self.stream.abort()
                self.start_stream(pitch)

    def on_playback_device_changed_signal(self):
        self.logger.debug(f"Playback device changed")
        self.stop_stream()
        self.start_stream()

    def stop_stream(self):
        if self.is_espeak:
            return
        if self.stream:
            self.logger.info("Stopping TTS vocalizer stream...")
            try:
                self.stream.stop()
            except Exception as e:
                self.logger.error(e)
            try:
                self.stream.close()
            except Exception as e:
                self.logger.error(e)
            try:
                self.stream.abort()
            except Exception as e:
                self.logger.error(e)
            self.stream = None

    def start_stream(self, pitch: Optional[int] = None):
        if self.is_espeak:
            return
        self.logger.info("Starting TTS vocalizer stream...")
        try:
            if sd.query_devices(kind="output"):
                if pitch is None and self.speech_t5_settings is not None:
                    pitch = self.speech_t5_settings.pitch
                else:
                    pitch = 100.0
                # set samplerate between 14000 and 24000
                # pitch == 0 -> samplerate == 14000
                # pitch == 50 -> samplerate == 19000
                # pitch == 100 -> samplerate == 24000
                samplerate = 14000 + int(10000.0 * (pitch / 100.0))
                self._initialize_stream(samplerate)
        except sd.PortAudioError as e:
            self.logger.error(f"Failed to start audio stream: {e}")
            self.stream = None

    @property
    def playback_device(self):
        playback_device = self.sound_settings.playback_device
        return playback_device if playback_device != "" else "pulse"

    @property
    def selected_device(self):
        if not self._selected_device:
            devices = sd.query_devices()
            for device in devices:
                if device["name"] == self.playback_device:
                    self._selected_device = device["index"]
                    break
        return self._selected_device

    def _initialize_stream(self, samplerate: int):
        self.logger.info(
            f"Initializing TTS stream with samplerate: {samplerate}"
        )
        try:
            if self.selected_device is not None:
                self.stream = sd.OutputStream(
                    samplerate=samplerate,
                    channels=1,
                    device=self.selected_device,
                )
                self.stream.start()
            else:
                self.logger.error("No valid playback device found.")
                self.stream = None
        except sd.PortAudioError as e:
            self.logger.error(f"Failed to start audio stream: {e}")
            self.stream = None

    def on_tts_generator_worker_add_to_stream_signal(self, response: dict):
        if self.accept_message:
            self.logger.debug("Adding speech to stream...")
            self.add_to_queue(response["message"])

    def handle_message(self, item):
        if not self.accept_message:
            return

        if item is None:
            self.logger.warning("item is none")
            return

        if self.stream is None:
            self.start_stream()
            self.logger.warning("No speakers available")
            return

        # Write the item to the stream
        if self.stream is not None and self.stream.active:
            try:
                self.stream.write(item)
            except sd.PortAudioError:
                self.logger.debug("PortAudioError")
            except AttributeError:
                self.logger.debug("stream is None")

            self.started = True
        QThread.msleep(AIRUNNER_SLEEP_TIME_IN_MS)

    def handle_speech(self, generated_speech):
        self.logger.debug("Adding speech to stream...")
        try:
            self.queue.put(generated_speech)
        except Exception as e:
            self.logger.error(f"Error while adding speech to stream: {e}")



import threading
import time
import sounddevice as sd
from airunner.gui.widgets.base_widget import BaseWidget
from airunner.gui.widgets.sound_settings.templates.sound_settings_ui import (
    Ui_SoundSettings,
)
from airunner.data.models import SoundSettings
from airunner.enums import SignalCode


class SoundSettingsWidget(BaseWidget):
    widget_class_ = Ui_SoundSettings

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.load_devices()  # Ensure devices are loaded on initialization
        self.connect_signals()
        self.monitoring = True
        # self.start_microphone_monitor()

    def load_devices(self):
        # Populate comboboxes with available audio devices
        devices = sd.query_devices()
        output_devices = filter(
            None,
            [
                None if device["max_output_channels"] == 0 else device
                for device in devices
            ],
        )
        input_devices = filter(
            None,
            [
                None if device["max_input_channels"] == 0 else device
                for device in devices
            ],
        )

        current_output_device = sd.query_devices(kind="output")
        current_input_device = sd.query_devices(kind="input")

        input_device_names = [device["name"] for device in input_devices]
        output_device_names = [device["name"] for device in output_devices]

        self.ui.playbackComboBox.clear()
        self.ui.playbackComboBox.addItems(output_device_names)
        self.ui.playbackComboBox.setCurrentText(current_output_device["name"])

        self.ui.recordingComboBox.clear()
        self.ui.recordingComboBox.addItems(input_device_names)
        self.ui.recordingComboBox.setCurrentText(current_input_device["name"])

        # Retrieve current settings
        sound_settings = SoundSettings.objects.first()

        # Set playback device
        if (
            sound_settings
            and sound_settings.playback_device in output_device_names
        ):
            self.ui.playbackComboBox.setCurrentText(
                sound_settings.playback_device
            )
        elif output_device_names:
            current_device = output_device_names[0]
            self.ui.playbackComboBox.setCurrentIndex(0)
            sound_settings = SoundSettings.objects.first()
            if sound_settings:
                SoundSettings.objects.update(
                    sound_settings.id, playback_device=current_device
                )

        # Set recording device
        if (
            sound_settings
            and sound_settings.recording_device in input_device_names
        ):
            self.ui.recordingComboBox.setCurrentText(
                sound_settings.recording_device
            )
        elif input_device_names:
            current_device = input_device_names[0]
            self.ui.recordingComboBox.setCurrentIndex(0)
            sound_settings = SoundSettings.objects.first()
            if sound_settings:
                SoundSettings.objects.update(
                    sound_settings.id, recording_device=current_device
                )

        self.ui.microphoneLevelBar.setValue(
            sound_settings.microphone_volume if sound_settings else 0
        )
        self.ui.inputLevelSlider.setValue(
            sound_settings.microphone_volume if sound_settings else 0
        )

    def connect_signals(self):
        self.ui.playbackComboBox.currentTextChanged.connect(
            self.update_playback_device
        )
        self.ui.recordingComboBox.currentTextChanged.connect(
            self.update_recording_device
        )
        self.ui.inputLevelSlider.valueChanged.connect(
            self.update_microphone_volume
        )

    def update_playback_device(self, device):
        sound_settings = SoundSettings.objects.first()
        SoundSettings.objects.update(sound_settings.id, playback_device=device)
        self.emit_signal(SignalCode.PLAYBACK_DEVICE_CHANGED, device)

    def update_recording_device(self, device):
        sound_settings = SoundSettings.objects.first()
        SoundSettings.objects.update(
            sound_settings.id, recording_device=device
        )
        self.emit_signal(SignalCode.RECORDING_DEVICE_CHANGED, device)

    def update_microphone_volume(self, volume):
        sound_settings = SoundSettings.objects.first()
        SoundSettings.objects.update(
            sound_settings.id, microphone_volume=volume
        )

    def adjust_input_level(self, value):
        # Adjust the microphone input level
        print(f"Input level adjusted to: {value}")
        # Replace with actual logic to adjust microphone input level

    def start_microphone_monitor(self):
        # Start a thread to monitor real-time microphone levels
        def monitor():
            while self.monitoring:
                try:
                    # Capture a short audio sample
                    audio_data = sd.rec(
                        int(0.1 * 44100),
                        samplerate=44100,
                        channels=1,
                        dtype="float32",
                    )
                    sd.wait()
                    # Calculate the RMS (Root Mean Square) level of the audio sample
                    rms_level = int((audio_data**2).mean() ** 0.5 * 100)
                    self.ui.microphoneLevelBar.setValue(min(rms_level, 100))
                except Exception as e:
                    print(f"Error monitoring microphone: {e}")
                time.sleep(0.1)

        threading.Thread(target=monitor, daemon=True).start()

    def closeEvent(self, event):
        # Stop monitoring when the widget is closed
        self.monitoring = False
        super().closeEvent(event)



from sqlalchemy import Column, Integer, String, Boolean, LargeBinary

from airunner.data.models.base import BaseModel


class ControlnetSettings(BaseModel):
    __tablename__ = 'controlnet_settings'
    id = Column(Integer, primary_key=True, autoincrement=True)
    image = Column(LargeBinary, nullable=True)
    generated_image = Column(LargeBinary, nullable=True)
    enabled = Column(Boolean, default=False)
    use_grid_image_as_input = Column(Boolean, default=False)
    strength = Column(Integer, default=50)
    conditioning_scale = Column(Integer, default=100)
    guidance_scale = Column(Integer, default=750)
    controlnet = Column(String, default="Canny")
    lock_input_image = Column(Boolean, default=False)



from airunner.settings import AIRUNNER_ART_ENABLED


controlnet_bootstrap_data = [
    {
        "display_name": "Canny",
        "name": "canny",
        "path": "lllyasviel/control_v11p_sd15_canny",
        "version": "SD 1.5",
        "pipeline_action": "controlnet",
        "size": "722600",
    },
    {
        "display_name": "Depth",
        "name": "depth_leres",
        "path": "lllyasviel/control_v11f1p_sd15_depth",
        "version": "SD 1.5",
        "pipeline_action": "controlnet",
        "size": "722600",
    },
    {
        "display_name": "MLSD",
        "name": "mlsd",
        "path": "lllyasviel/control_v11p_sd15_mlsd",
        "version": "SD 1.5",
        "pipeline_action": "controlnet",
        "size": "722600",
    },
    {
        "display_name": "Normal",
        "name": "normal_bae",
        "path": "lllyasviel/control_v11p_sd15_normalbae",
        "version": "SD 1.5",
        "pipeline_action": "controlnet",
        "size": "722600",
    },
    {
        "display_name": "Scribble",
        "name": "scribble_hed",
        "path": "lllyasviel/control_v11p_sd15_scribble",
        "version": "SD 1.5",
        "pipeline_action": "controlnet",
        "size": "722600",
    },
    {
        "display_name": "Segmentation",
        "name": "segmentation",
        "path": "lllyasviel/control_v11p_sd15_seg",
        "version": "SD 1.5",
        "pipeline_action": "controlnet",
        "size": "722600",
    },
    {
        "display_name": "Lineart",
        "name": "lineart_coarse",
        "path": "lllyasviel/control_v11p_sd15_lineart",
        "version": "SD 1.5",
        "pipeline_action": "controlnet",
        "size": "722600",
    },
    {
        "display_name": "Lineart Anime",
        "name": "lineart_anime",
        "path": "lllyasviel/control_v11p_sd15s2_lineart_anime",
        "version": "SD 1.5",
        "pipeline_action": "controlnet",
        "size": "722600",
    },
    {
        "display_name": "Openpose",
        "name": "openpose",
        "path": "lllyasviel/control_v11p_sd15_openpose",
        "version": "SD 1.5",
        "pipeline_action": "controlnet",
        "size": "722600",
    },
    {
        "display_name": "Softedge",
        "name": "softedge_hed",
        "path": "lllyasviel/control_v11p_sd15_softedge",
        "version": "SD 1.5",
        "pipeline_action": "controlnet",
        "size": "722600",
    },
    {
        "display_name": "Pixel2Pixel",
        "name": "pixel2pixel",
        "path": "lllyasviel/control_v11e_sd15_ip2p",
        "version": "SD 1.5",
        "pipeline_action": "controlnet",
        "size": "722600",
    },
    {
        "display_name": "Inpaint",
        "name": "inpaint",
        "path": "lllyasviel/control_v11p_sd15_inpaint",
        "version": "SD 1.5",
        "pipeline_action": "controlnet",
        "size": "722600",
    },
    {
        "display_name": "Shuffle",
        "name": "shuffle",
        "path": "lllyasviel/control_v11e_sd15_shuffle",
        "version": "SD 1.5",
        "pipeline_action": "controlnet",
        "size": "722600",
    },
    {
        "display_name": "Canny SDXL",
        "name": "canny",
        "path": "diffusers/controlnet-canny-sdxl-1.0-small",
        "version": "SDXL 1.0",
        "pipeline_action": "controlnet",
        "size": "320200",
    },
]


if not AIRUNNER_ART_ENABLED:
    controlnet_bootstrap_data = []




from PySide6.QtGui import QPolygonF, QPen, QBrush
from PySide6.QtCore import QPointF, Qt

from NodeGraphQt import BaseNode
from NodeGraphQt.constants import PortTypeEnum

from airunner.utils.application.mediator_mixin import MediatorMixin
from airunner.gui.windows.main.settings_mixin import SettingsMixin


class BaseWorkflowNode(
    MediatorMixin,
    SettingsMixin,
    BaseNode,
):
    # Base identifier for easier registration and type checking
    __identifier__ = "airunner.workflow.nodes.BaseWorkflowNode"
    # Execution port constants
    EXEC_IN_PORT_NAME = "exec_in"
    EXEC_OUT_PORT_NAME = "exec_out"
    has_exec_in_port: bool = True
    has_exec_out_port: bool = True

    def __init__(self):
        super().__init__()
        # Add standard execution ports
        # Allow only one connection for execution ports (1-to-1)
        if self.has_exec_in_port:
            self.add_input(
                self.EXEC_IN_PORT_NAME,
                multi_input=False,  # Only one connection allowed to exec_in
                display_name=False,
                painter_func=self._draw_exec_port,
            )
        if self.has_exec_out_port:
            self.add_output(
                self.EXEC_OUT_PORT_NAME,
                multi_output=False,  # Only one connection allowed from exec_out
                display_name=False,
                painter_func=self._draw_exec_port,
            )

        # Connect to the connection changed signals
        if hasattr(self.graph, "connection_changed"):
            self.graph.connection_changed.connect(self._on_connection_changed)

    def on_input_connected(self, in_port, out_port):
        """
        Override method called when a connection is made to an input port.
        Ensures exec_in port only maintains one connection.
        """
        super().on_input_connected(in_port, out_port)
        if in_port.name() == self.EXEC_IN_PORT_NAME:
            # Disconnect any other connections to this exec_in port except the new one
            for connected_port in in_port.connected_ports():
                if connected_port != out_port:
                    in_port.disconnect_from(connected_port)

    def on_output_connected(self, out_port, in_port):
        """
        Override method called when a connection is made from an output port.
        Ensures exec_out port only maintains one connection.
        """
        super().on_output_connected(out_port, in_port)
        if out_port.name() == self.EXEC_OUT_PORT_NAME:
            # Disconnect any other connections from this exec_out port except the new one
            for connected_port in out_port.connected_ports():
                if connected_port != in_port:
                    out_port.disconnect_from(connected_port)

    def _on_connection_changed(self, disconnected, connected):
        """
        Handle connection changes in the graph.
        This is a fallback method if the on_input_connected and on_output_connected
        are not sufficient.
        """
        if not connected:
            return

        # For each new connection, check if it involves our exec ports
        for src_port, tgt_port in connected:
            # If this node is the source with exec_out
            if (
                src_port.node().id() == self.id()
                and src_port.name() == self.EXEC_OUT_PORT_NAME
            ):
                out_port = self.output(self.EXEC_OUT_PORT_NAME)
                # Disconnect all others except the new connection
                for conn_port in list(out_port.connected_ports()):
                    if conn_port != tgt_port:
                        out_port.disconnect_from(conn_port)

            # If this node is the target with exec_in
            elif (
                tgt_port.node().id() == self.id()
                and tgt_port.name() == self.EXEC_IN_PORT_NAME
            ):
                in_port = self.input(self.EXEC_IN_PORT_NAME)
                # Disconnect all others except the new connection
                for conn_port in list(in_port.connected_ports()):
                    if conn_port != src_port:
                        in_port.disconnect_from(conn_port)

    # Custom painter function for execution ports (simple triangle)
    @staticmethod
    def _draw_exec_port(painter, rect, info):
        painter.setPen(Qt.PenStyle.NoPen)
        painter.setBrush(QBrush(Qt.GlobalColor.white))
        points = [
            QPointF(rect.left(), rect.top()),
            QPointF(rect.right(), rect.center().y()),
            QPointF(rect.left(), rect.bottom()),
        ]
        painter.drawPolygon(QPolygonF(points))

    # Method to dynamically add an input port
    def add_dynamic_input(
        self,
        name="input",
        multi_input=False,
        display_name=True,
        color=None,
    ):
        # Ensure port name is unique before adding
        if name not in self.inputs():
            return self.add_input(
                name,
                multi_input=multi_input,
                display_name=display_name,
                color=color,
            )
        return self.input(name)

    # Method to dynamically add an output port
    def add_dynamic_output(
        self,
        name="output",
        multi_output=True,
        display_name=True,
        color=None,
    ):
        # Ensure port name is unique before adding
        if name not in self.outputs():
            return self.add_output(
                name,
                multi_output=multi_output,
                display_name=display_name,
                color=color,
            )
        return self.output(name)

    # Placeholder for execution logic - subclasses should override
    def execute(self, input_data):
        print(
            f"Executing node {self.name()} - Base implementation passes data through."
        )
        # By default, pass input data through if an output exists
        passthrough_data = {}
        # Find the first non-exec output port to pass data through
        data_output_port_name = next(
            (
                name
                for name, port in self.outputs().items()
                if name != self.EXEC_OUT_PORT_NAME
            ),
            None,
        )

        # Find the first non-exec input port to get data from
        data_input_port_name = next(
            (
                name
                for name, port in self.inputs().items()
                if name != self.EXEC_IN_PORT_NAME
            ),
            None,
        )

        if (
            data_output_port_name
            and data_input_port_name
            and data_input_port_name in input_data
        ):
            passthrough_data[data_output_port_name] = input_data[
                data_input_port_name
            ]

        # Return data keyed by output port name and indicate execution flow
        return {**passthrough_data, "_exec_triggered": self.EXEC_OUT_PORT_NAME}

    # Helper to get data from a specific input port name
    def get_input_data(self, port_name, input_data, default=None):
        return input_data.get(port_name, default)



"""
RabbitMQ backend for SignalMediator.

This requires the pika library to be installed.

"""

from typing import Callable, Dict
from airunner.enums import SignalCode


class RabbitMQBackend:
    """
    RabbitMQ-based backend for SignalMediator.
    """

    def __init__(self, username: str, password: str, url: str, port: int):
        import pika

        rabbitmq_url = f"amqp://{username}:{password}@{url}:{port}/"
        self.rabbitmq_url = rabbitmq_url
        self.connection = pika.BlockingConnection(
            pika.URLParameters(rabbitmq_url)
        )
        self.channel = self.connection.channel()
        self.handlers: Dict[str, Callable] = {}

    def register(self, code: SignalCode, handler: Callable):
        """
        Register a handler for a specific SignalCode.
        """
        queue_name = code.value
        self.handlers[queue_name] = handler
        self.channel.queue_declare(queue=queue_name)

        def callback(ch, method, properties, body):
            handler(body)

        self.channel.basic_consume(
            queue=queue_name, on_message_callback=callback, auto_ack=True
        )

    def emit_signal(self, code: SignalCode, data: dict):
        """
        Publish a message to the RabbitMQ queue.
        """
        queue_name = code.value
        self.channel.queue_declare(queue=queue_name)
        self.channel.basic_publish(
            exchange="", routing_key=queue_name, body=str(data)
        )

    def start_consuming(self):
        """
        Start consuming messages from RabbitMQ.
        """
        self.channel.start_consuming()

    def close(self):
        """
        Close the RabbitMQ connection.
        """
        self.connection.close()



from typing import Dict, List, Any, Type

import diffusers
from diffusers.pipelines.stable_diffusion.pipeline_stable_diffusion import (
    StableDiffusionPipeline,
)
from diffusers.pipelines.stable_diffusion.pipeline_stable_diffusion_img2img import (
    StableDiffusionImg2ImgPipeline,
)
from diffusers.pipelines.stable_diffusion.pipeline_stable_diffusion_inpaint import (
    StableDiffusionInpaintPipeline,
)
from diffusers.pipelines.controlnet.pipeline_controlnet import (
    StableDiffusionControlNetPipeline,
)
from diffusers.pipelines.controlnet.pipeline_controlnet_img2img import (
    StableDiffusionControlNetImg2ImgPipeline,
)
from diffusers.pipelines.controlnet.pipeline_controlnet_inpaint import (
    StableDiffusionControlNetInpaintPipeline,
)
from airunner.handlers.stablediffusion.base_diffusers_model_manager import (
    BaseDiffusersModelManager,
)


class StableDiffusionModelManager(BaseDiffusersModelManager):
    @property
    def img2img_pipelines(self):
        return (
            StableDiffusionImg2ImgPipeline,
            StableDiffusionControlNetImg2ImgPipeline,
        )

    @property
    def txt2img_pipelines(self):
        return (
            StableDiffusionPipeline,
            StableDiffusionControlNetPipeline,
        )

    @property
    def controlnet_pipelines(self):
        return (
            StableDiffusionControlNetPipeline,
            StableDiffusionControlNetImg2ImgPipeline,
            StableDiffusionControlNetInpaintPipeline,
        )

    @property
    def outpaint_pipelines(self):
        return (
            StableDiffusionInpaintPipeline,
            StableDiffusionControlNetInpaintPipeline,
        )

    @property
    def pipeline_map(
        self,
    ) -> Dict[str, Type[diffusers.pipelines.pipeline_utils.DiffusionPipeline]]:
        return {
            "txt2img": StableDiffusionPipeline,
            "img2img": StableDiffusionImg2ImgPipeline,
            "outpaint": StableDiffusionInpaintPipeline,
            "txt2img_controlnet": StableDiffusionControlNetPipeline,
            "img2img_controlnet": StableDiffusionControlNetImg2ImgPipeline,
            "outpaint_controlnet": StableDiffusionControlNetInpaintPipeline,
        }

    @property
    def _pipeline_class(
        self,
    ) -> Type[diffusers.pipelines.pipeline_utils.DiffusionPipeline]:
        operation_type = "txt2img"
        if self.is_img2img:
            operation_type = "img2img"
        elif self.is_outpaint:
            operation_type = "outpaint"

        if self.controlnet_enabled:
            operation_type = f"{operation_type}_controlnet"

        return self.pipeline_map.get(operation_type)

    @property
    def second_prompt(self) -> str:
        prompt = self.image_request.second_prompt
        prompt_preset = self.prompt_preset

        # Format the prompt
        formatted_prompt = None
        if self.do_join_prompts:
            prompts = [f'"{prompt}"']
            for (
                additional_prompt_settings
            ) in self.image_request.additional_prompts:
                addtional_prompt = additional_prompt_settings[
                    "prompt_secondary"
                ]
                prompts.append(f'"{addtional_prompt}"')
            formatted_prompt = (
                f'({", ".join(prompts)}, "{prompt_preset}").and()'
            )

        if prompt_preset != "":
            prompt = f'("{prompt}", "{prompt_preset}").and()'

        formatted_prompt = formatted_prompt or prompt

        return formatted_prompt

    @property
    def second_negative_prompt(self) -> str:
        prompt = self.image_request.second_negative_prompt
        negative_prompt_preset = self.negative_prompt_preset

        if negative_prompt_preset != "":
            prompt = f'("{prompt}", "{negative_prompt_preset}").and()'

        return prompt

    @property
    def compel_tokenizer(self) -> Any:
        return self._pipe.tokenizer

    @property
    def compel_text_encoder(self) -> Any:
        return self._pipe.text_encoder



import os
from typing import Dict, List, Type, Any

import diffusers
from diffusers.pipelines.stable_diffusion_xl.pipeline_stable_diffusion_xl import (
    StableDiffusionXLPipeline,
)
from diffusers.pipelines.stable_diffusion_xl.pipeline_stable_diffusion_xl_img2img import (
    StableDiffusionXLImg2ImgPipeline,
)
from diffusers.pipelines.stable_diffusion_xl.pipeline_stable_diffusion_xl_inpaint import (
    StableDiffusionXLInpaintPipeline,
)
from diffusers.pipelines.controlnet.pipeline_controlnet_sd_xl import (
    StableDiffusionXLControlNetPipeline,
)
from diffusers.pipelines.controlnet.pipeline_controlnet_sd_xl_img2img import (
    StableDiffusionXLControlNetImg2ImgPipeline,
)
from diffusers.pipelines.controlnet.pipeline_controlnet_inpaint_sd_xl import (
    StableDiffusionXLControlNetInpaintPipeline,
)

from compel import ReturnedEmbeddingsType

from airunner.enums import StableDiffusionVersion
from airunner.handlers.stablediffusion.stable_diffusion_model_manager import (
    StableDiffusionModelManager,
)


class SDXLModelManager(StableDiffusionModelManager):
    @property
    def img2img_pipelines(self):
        return (
            StableDiffusionXLImg2ImgPipeline,
            StableDiffusionXLControlNetImg2ImgPipeline,
        )

    @property
    def txt2img_pipelines(self):
        return (
            StableDiffusionXLPipeline,
            StableDiffusionXLControlNetPipeline,
        )

    @property
    def controlnet_pipelines(self):
        return (
            StableDiffusionXLControlNetPipeline,
            StableDiffusionXLControlNetImg2ImgPipeline,
            StableDiffusionXLControlNetInpaintPipeline,
        )

    @property
    def outpaint_pipelines(self):
        return (
            StableDiffusionXLInpaintPipeline,
            StableDiffusionXLControlNetInpaintPipeline,
        )

    @property
    def is_sd_xl(self) -> bool:
        return self.real_model_version == StableDiffusionVersion.SDXL1_0.value

    @property
    def is_sd_xl_turbo(self) -> bool:
        return (
            self.real_model_version == StableDiffusionVersion.SDXL_TURBO.value
        )

    @property
    def is_sd_xl_or_turbo(self) -> bool:
        return self.is_sd_xl or self.is_sd_xl_turbo

    @property
    def version(self) -> str:
        """
        Turbo paths are SDXL 1.0 paths so we normalize the version to SDXL 1.0
        """
        version = super().version
        if version == "SDXL Turbo":
            version = "SDXL 1.0"
        return version

    @property
    def config_path(self) -> str:
        config_path = super().config_path
        if self.is_sd_xl_turbo:
            config_path = os.path.expanduser(
                os.path.join(
                    self.path_settings_cached.base_path,
                    "art",
                    "models",
                    StableDiffusionVersion.SDXL1_0.value,
                    self.image_request.pipeline_action,
                )
            )
        return config_path

    @property
    def pipeline_map(
        self,
    ) -> Dict[str, Type[diffusers.pipelines.pipeline_utils.DiffusionPipeline]]:
        return {
            "txt2img": StableDiffusionXLPipeline,
            "img2img": StableDiffusionXLImg2ImgPipeline,
            "outpaint": StableDiffusionXLInpaintPipeline,
            "txt2img_controlnet": StableDiffusionXLControlNetPipeline,
            "img2img_controlnet": StableDiffusionXLControlNetImg2ImgPipeline,
            "outpaint_controlnet": StableDiffusionXLControlNetInpaintPipeline,
        }

    @property
    def compel_parameters(self) -> Dict[str, Any]:
        parameters = super().compel_parameters
        parameters["returned_embeddings_type"] = (
            ReturnedEmbeddingsType.PENULTIMATE_HIDDEN_STATES_NON_NORMALIZED
        )
        parameters["requires_pooled"] = [False, True]
        return parameters

    @property
    def compel_tokenizer(self) -> List[Any]:
        return [self._pipe.tokenizer, self._pipe.tokenizer_2]

    @property
    def compel_text_encoder(self) -> List[Any]:
        return [self._pipe.text_encoder, self._pipe.text_encoder_2]

    def _prepare_compel_data(self, data: Dict) -> Dict:
        data = super()._prepare_compel_data(data)
        data.update(
            {
                "pooled_prompt_embeds": self._pooled_prompt_embeds,
                "negative_pooled_prompt_embeds": self._negative_pooled_prompt_embeds,
            }
        )

        for key in [
            "negative_target_size",
            "negative_original_size",
            "crops_coords_top_left",
        ]:
            val = getattr(self.image_request, key, None)
            if val and val.get("width", None) and val.get("height", None):
                data.update({key: (val["width"], val["height"])})
        return data

    def _build_conditioning_tensors(
        self, compel_prompt, compel_negative_prompt
    ):
        prompt_embeds, pooled_prompt_embeds = (
            self._compel_proc.build_conditioning_tensor(compel_prompt)
        )
        negative_prompt_embeds, negative_pooled_prompt_embeds = (
            self._compel_proc.build_conditioning_tensor(compel_negative_prompt)
        )
        return (
            prompt_embeds,
            pooled_prompt_embeds,
            negative_prompt_embeds,
            negative_pooled_prompt_embeds,
        )



from PySide6.QtUiTools import QUiLoader
from PySide6.QtCore import QFile, QBuffer, QIODevice, QObject
from PySide6.QtWidgets import QWidget

def load_ui_file(ui_file_path: str, parent: QWidget = None) -> QWidget:
    """
    Load a .ui file dynamically at runtime and return the corresponding QWidget.

    :param ui_file_path: Path to the .ui file.
    :param parent: Optional parent widget.
    :return: QWidget instance loaded from the .ui file.
    """
    loader = QUiLoader()
    ui_file = QFile(ui_file_path)
    if not ui_file.open(QFile.ReadOnly):
        raise FileNotFoundError(f"Unable to open UI file: {ui_file_path}")

    widget = loader.load(ui_file, parent)
    ui_file.close()

    if widget is None:
        raise RuntimeError(f"Failed to load UI file: {ui_file_path}")

    return widget

def load_ui_from_string(ui_content: str, parent: QWidget = None, signal_handler: QObject = None) -> QWidget:
    """
    Load a .ui file dynamically from a string and return the corresponding QWidget.

    :param ui_content: The content of the .ui file as a string.
    :param parent: Optional parent widget.
    :param signal_handler: Optional object to handle signals.
    :return: QWidget instance loaded from the .ui content.
    """
    loader = QUiLoader()
    buffer = QBuffer()
    buffer.setData(ui_content.encode('utf-8'))
    if not buffer.open(QIODevice.ReadOnly):
        raise RuntimeError("Unable to open UI content from string.")

    widget = loader.load(buffer, parent)
    buffer.close()

    if widget is None:
        raise RuntimeError("Failed to load UI content from string.")

    # Automatically connect signals if a signal handler is provided
    if signal_handler:
        for child in widget.findChildren(QObject):
            if hasattr(signal_handler, child.objectName()):
                signal = getattr(signal_handler, child.objectName())
                if callable(signal):
                    child.clicked.connect(signal)

    return widget


from sqlalchemy import Column, Integer, String

from airunner.data.models.base import BaseModel


class VoiceSettings(BaseModel):
    __tablename__ = "voice_settings"
    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, nullable=False, unique=True)
    model_type = Column(
        String, nullable=False
    )  # TTSModel string
    settings_id = Column(
        Integer, nullable=False
    )  # Links to EspeakSettings or SpeechT5Settings or OpenVoiceSettings




from PySide6.QtWidgets import QColorDialog
from airunner.enums import SignalCode
from airunner.gui.widgets.active_grid_settings.templates.active_grid_settings_ui import Ui_active_grid_settings_widget
from airunner.gui.widgets.base_widget import BaseWidget
from PySide6.QtCore import Slot


class ActiveGridSettingsWidget(BaseWidget):
    widget_class_ = Ui_active_grid_settings_widget

    def __init__(self, *args, **kwargs):
        self.signal_handlers = {
            SignalCode.APPLICATION_SETTINGS_CHANGED_SIGNAL: self.update_size,
            SignalCode.APPLICATION_ACTIVE_GRID_AREA_UPDATED: self.update_size
        }
        super().__init__(*args, **kwargs)
        self.ui.width_slider_widget.setProperty("current_value", self.application_settings.is_maximized)
        self.ui.height_slider_widget.setProperty("current_value", self.application_settings.working_height)
        self.ui.width_slider_widget.initialize()
        self.ui.height_slider_widget.initialize()
        self.ui.border_opacity_slider_widget.setProperty(
            "current_value",
            self.active_grid_settings.border_opacity
        )
        self.ui.fill_opacity_slider_widget.setProperty(
            "current_value",
            self.active_grid_settings.fill_opacity
        )

        self.ui.border_opacity_slider_widget.initialize()
        self.ui.fill_opacity_slider_widget.initialize()
        self.ui.active_grid_border_groupbox.blockSignals(True)
        self.ui.active_grid_fill_groupbox.blockSignals(True)
        self.ui.active_grid_area_checkbox.blockSignals(True)
        self.ui.size_lock_button.blockSignals(True)
        self.ui.active_grid_border_groupbox.setChecked(self.active_grid_settings.render_border)
        self.ui.active_grid_fill_groupbox.setChecked(self.active_grid_settings.render_fill)
        self.ui.active_grid_area_checkbox.setChecked(self.active_grid_settings.enabled)
        self.ui.size_lock_button.setChecked(self.application_settings.active_grid_size_lock)
        self.ui.active_grid_area_checkbox.blockSignals(False)
        self.ui.active_grid_border_groupbox.blockSignals(False)
        self.ui.active_grid_fill_groupbox.blockSignals(False)
        self.ui.size_lock_button.blockSignals(False)

        # set background color of buttons
        self.ui.border_choose_color_button.setStyleSheet(
            f"background-color: {self.active_grid_settings.border_color}"
        )
        self.ui.fill_choose_color_button.setStyleSheet(
            f"background-color: {self.active_grid_settings.fill_color}"
        )
        self.current_active_grid_width = self.application_settings.working_width
        self.current_active_grid_height = self.application_settings.working_height

    @Slot(bool)
    def size_lock_toggled(self, val):
        self.update_application_settings("active_grid_size_lock", val)

    def update_size(self, _message: dict):
        width = self.application_settings.working_width
        height = self.application_settings.working_height
        if self.application_settings.active_grid_size_lock:
            if width != self.current_active_grid_width:
                height = width
            elif height != self.current_active_grid_height:
                width = height
        if self.current_active_grid_height != height or self.current_active_grid_width != width:
            self.current_active_grid_width = width
            self.current_active_grid_height = height
            self.ui.width_slider_widget.blockSignals(True)
            self.ui.height_slider_widget.blockSignals(True)
            self.ui.width_slider_widget.set_slider_and_spinbox_values(width)
            self.ui.height_slider_widget.set_slider_and_spinbox_values(height)
            self.ui.width_slider_widget.blockSignals(False)
            self.ui.height_slider_widget.blockSignals(False)
            self.update_application_settings("working_width", width)
            self.update_application_settings("working_height", height)

    def action_clicked_checkbox_toggle_active_grid_border(self, checked):
        self.update_active_grid_settings("render_border", checked)

    def action_clicked_checkbox_toggle_active_grid_fill(self, checked):
        self.update_active_grid_settings("render_fill", checked)

    def action_clicked_checkbox_toggle_active_grid_area(self, checked):
        self.update_active_grid_settings("enabled", checked)

    def action_choose_border_color_clicked(self):
        color = QColorDialog.getColor()
        if color.isValid():
            self.update_active_grid_settings("border_color", color.name())

    def action_choose_fill_color_clicked(self):
        color = QColorDialog.getColor()
        if color.isValid():
            self.update_active_grid_settings("fill_color", color.name())



import os

from airunner.data.models import AIModels
from airunner.enums import SignalCode
from airunner.gui.windows.main.pipeline_mixin import PipelineMixin
from airunner.workers.worker import Worker


class ModelScannerWorker(Worker, PipelineMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        PipelineMixin.__init__(self)

    def handle_message(self, _message):
        self.scan_for_models()
        self.remove_missing_models()

    def scan_for_models(self):
        self.logger.debug("Scan for models")
        # look at model path and determine if we can import existing local models
        # first look at all files and folders inside of the model paths
        diffusers_folders = [
            "scheduler",
            "text_encoder",
            "tokenizer",
            "unet",
            "vae",
        ]
        models = []
        model_path = os.path.expanduser(
            os.path.join(
                self.path_settings.base_path,
                "art/models",
            )
        )
        if not os.path.exists(model_path):
            self.logger.error(f"Model path does not exist: {model_path}")
            return
        # find all folders inside of model_path, each of those folders is a model version
        with os.scandir(model_path) as dir_object:
            # check if dir_object is a directory
            self.logger.debug(f"Scan for models {model_path}")
            for version_entry in dir_object:
                version = version_entry.name
                path = os.path.join(model_path, version)
                self.logger.debug(f"Scan directory {path}")
                with os.scandir(path) as action_object:
                    for action_item in action_object:
                        action = action_item.name
                        if "controlnet_processors" in action_item.path:
                            continue
                        paths = [action_item.path]
                        if "SDXL 1.0/txt2img" in action_item.path:
                            paths.append(
                                os.path.join(action_item.path, "turbo_models")
                            )
                        for path in paths:
                            if not os.path.exists(path):
                                continue
                            with os.scandir(path) as file_object:
                                for file_item in file_object:
                                    model = AIModels()
                                    model.name = os.path.basename(
                                        file_item.path
                                    )
                                    model.path = file_item.path
                                    model.branch = "main"
                                    if "turbo_models" in path:
                                        version = "SDXL Turbo"
                                    model.version = version
                                    model.category = "stablediffusion"
                                    model.pipeline_action = action
                                    model.enabled = True
                                    model.model_type = "art"
                                    model.is_default = False
                                    if (
                                        file_item.is_file()
                                    ):  # ckpt or safetensors file
                                        if file_item.name.endswith(
                                            ".ckpt"
                                        ) or file_item.name.endswith(
                                            ".safetensors"
                                        ):
                                            name = file_item.name.replace(
                                                ".ckpt", ""
                                            ).replace(".safetensors", "")
                                            model.name = name
                                        else:
                                            model = None
                                    elif (
                                        file_item.is_dir()
                                    ):  # diffusers folder
                                        is_diffusers_directory = True
                                        for (
                                            diffuser_folder
                                        ) in diffusers_folders:
                                            if not os.path.exists(
                                                os.path.join(
                                                    file_item.path,
                                                    diffuser_folder,
                                                )
                                            ):
                                                is_diffusers_directory = False
                                                model = None
                                        if is_diffusers_directory:
                                            model.name = file_item.name

                                    if model:
                                        models.append(model)
        self.emit_signal(
            SignalCode.AI_MODELS_SAVE_OR_UPDATE_SIGNAL, {"models": models}
        )

    def remove_missing_models(self):
        # remove all models that are not in the model path
        model_path = os.path.expanduser(
            os.path.join(
                self.path_settings.base_path,
                "art/models",
            )
        )
        if not os.path.exists(model_path):
            self.logger.error(f"Model path does not exist: {model_path}")
            return
        existing_models = AIModels.objects.all()
        for model in existing_models:
            if not os.path.exists(model.path):
                self.logger.debug(f"Remove missing model {model.id}")
                AIModels.objects.delete(model.id)
                self.logger.debug(f"Removed missing model: {model.name}")



from PySide6.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QLineEdit,
    QComboBox,
    QPushButton,
    QMessageBox,
)
from airunner.gui.widgets.tts.templates.voice_settings_ui import (
    Ui_voice_settings,
)
from airunner.data.models.voice_settings import VoiceSettings
from airunner.data.models.espeak_settings import EspeakSettings
from airunner.data.models.speech_t5_settings import SpeechT5Settings
from airunner.gui.widgets.tts.espeak_preferences_widget import (
    EspeakPreferencesWidget,
)
from airunner.gui.widgets.tts.speecht5_preferences_widget import (
    SpeechT5PreferencesWidget,
)
from airunner.settings import AIRUNNER_ENABLE_OPEN_VOICE
from airunner.gui.widgets.tts.open_voice_preferences_widget import (
    OpenVoicePreferencesWidget,
)
from airunner.enums import TTSModel
from airunner.data.models.openvoice_settings import OpenVoiceSettings


class VoiceSettingsWidget(QWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ui = Ui_voice_settings()
        self.ui.setupUi(self)
        self.ui.create_voice_button.clicked.connect(self.create_voice)
        self.load_voices()

    def load_voices(self):
        layout = self.ui.voice_list_layout
        while layout.count():
            item = layout.takeAt(0)
            widget = item.widget()
            if widget:
                widget.deleteLater()

        voices = VoiceSettings.objects.all()
        for voice in voices:
            self.add_voice_item(voice)

    def add_voice_item(self, voice):
        container = QWidget()
        container_layout = QVBoxLayout(container)

        name_edit = QLineEdit(voice.name)
        name_edit.textChanged.connect(
            lambda val, v=voice: self.update_voice_name(v, val)
        )
        container_layout.addWidget(name_edit)

        model_combobox = QComboBox()
        model_combobox.addItems(
            [TTSModel.SPEECHT5.value, TTSModel.ESPEAK.value]
        )
        if AIRUNNER_ENABLE_OPEN_VOICE:
            model_combobox.addItem(TTSModel.OPENVOICE.value)
        model_combobox.setCurrentText(voice.model_type)
        model_combobox.currentTextChanged.connect(
            lambda val, v=voice: self.update_voice_model(
                v, val, container_layout
            )
        )
        container_layout.addWidget(model_combobox)

        delete_button = QPushButton("Delete")
        delete_button.clicked.connect(lambda _, v=voice: self.delete_voice(v))
        container_layout.addWidget(delete_button)

        # Add the corresponding settings widget
        self.add_settings_widget(voice, container_layout)

        self.ui.voice_list_layout.addWidget(container)

    def add_settings_widget(self, voice, layout):
        """Add the appropriate settings widget (Espeak, SpeechT5, or OpenVoice) to the layout."""
        # Remove any existing settings widget
        for i in reversed(range(layout.count())):
            widget = layout.itemAt(i).widget()
            if isinstance(
                widget,
                (
                    EspeakPreferencesWidget,
                    SpeechT5PreferencesWidget,
                    OpenVoicePreferencesWidget,
                ),
            ):
                widget.deleteLater()

        # Add the appropriate settings widget
        if voice.model_type == TTSModel.SPEECHT5.value:
            widget = SpeechT5PreferencesWidget(id=voice.settings_id)
        elif voice.model_type == TTSModel.ESPEAK.value:
            widget = EspeakPreferencesWidget(id=voice.settings_id)
        elif (
            AIRUNNER_ENABLE_OPEN_VOICE
            and voice.model_type == TTSModel.OPENVOICE.value
        ):
            widget = OpenVoicePreferencesWidget(id=voice.settings_id)
        else:
            return

        layout.addWidget(widget)

    def create_voice(self):
        voice = VoiceSettings.objects.create(
            name="New Voice",
            model_type=(
                TTSModel.SPEECHT5.value
                if not AIRUNNER_ENABLE_OPEN_VOICE
                else TTSModel.OPENVOICE.value
            ),
            settings_id=0,
        )
        if voice.model_type == TTSModel.SPEECHT5.value:
            settings = SpeechT5Settings.objects.create()
        elif voice.model_type == TTSModel.ESPEAK.value:
            settings = EspeakSettings.objects.create()
        elif (
            AIRUNNER_ENABLE_OPEN_VOICE
            and voice.model_type == TTSModel.OPENVOICE.value
        ):
            settings = OpenVoiceSettings.objects.create()
        else:
            return
        VoiceSettings.objects.update(
            voice.id,
            settings_id=settings.id,
        )
        voice.settings_id = settings.id
        self.add_voice_item(voice)

    def update_voice_name(self, voice, name):
        voice.name = name
        VoiceSettings.objects.update(voice.id, name=name)

    def update_voice_model(self, voice, model_type, layout):
        if voice.model_type != model_type:
            # Delete the old settings
            if voice.model_type == TTSModel.SPEECHT5:
                SpeechT5Settings.objects.delete(voice.settings_id)
            elif voice.model_type == TTSModel.ESPEAK:
                EspeakSettings.objects.delete(voice.settings_id)
            elif (
                AIRUNNER_ENABLE_OPEN_VOICE
                and voice.model_type == TTSModel.OPENVOICE
            ):
                OpenVoiceSettings.objects.delete(voice.settings_id)

            # Create new settings
            if model_type == TTSModel.SPEECHT5.value:
                settings = SpeechT5Settings.objects.create()
            elif model_type == TTSModel.ESPEAK.value:
                settings = EspeakSettings.objects.create()
            elif (
                AIRUNNER_ENABLE_OPEN_VOICE
                and model_type == TTSModel.OPENVOICE.value
            ):
                settings = OpenVoiceSettings.objects.create()
            else:
                return

            voice.settings_id = settings.id
            voice.model_type = model_type
            VoiceSettings.objects.update(
                voice.id, settings_id=settings.id, model_type=model_type
            )

        # Update the settings widget
        self.add_settings_widget(voice, layout)

    def delete_voice(self, voice):
        confirm = QMessageBox.question(
            self,
            "Delete Voice",
            f"Are you sure you want to delete '{voice.name}'?",
        )
        if confirm == QMessageBox.Yes:
            # Delete the associated settings
            if voice.model_type == TTSModel.SPEECHT5.value:
                SpeechT5Settings.objects.delete(voice.settings_id)
            elif voice.model_type == TTSModel.ESPEAK.value:
                EspeakSettings.objects.delete(voice.settings_id)
            elif (
                AIRUNNER_ENABLE_OPEN_VOICE
                and voice.model_type == TTSModel.OPENVOICE.value
            ):
                OpenVoiceSettings.objects.delete(voice.settings_id)

            VoiceSettings.objects.delete(voice.id)
            self.load_voices()
